"""Notifications."""

from __future__ import annotations

import sys
import threading
import warnings
from collections.abc import Sequence
from datetime import datetime
from enum import auto
from types import TracebackType
from typing import Callable, List, Optional, Tuple, Type, Union

from loguru import logger
from napari.utils.misc import StringEnum

try:
    from napari.utils.events import Event, EventEmitter
except ImportError:
    raise ImportError("please install napari using 'pip install napari'") from None

name2num = {
    "critical": 50,
    "error": 40,
    "warning": 30,
    "info": 20,
    "success": 20,
    "debug": 10,
    "none": 0,
}


class NotificationSeverity(StringEnum):
    """Severity levels for the notification dialog.  Along with icons for each."""

    NONE = auto()
    DEBUG = auto()
    INFO = auto()
    SUCCESS = auto()
    WARNING = auto()
    ERROR = auto()
    CRITICAL = auto()

    def __lt__(self, other):
        return name2num[str(self)] < name2num[str(other)]

    def __le__(self, other):
        return name2num[str(self)] <= name2num[str(other)]

    def __gt__(self, other):
        return name2num[str(self)] > name2num[str(other)]

    def __ge__(self, other):
        return name2num[str(self)] >= name2num[str(other)]


NOTIFICATION_LEVELS = {
    NotificationSeverity.NONE: "None",
    NotificationSeverity.DEBUG: "Debug",
    NotificationSeverity.INFO: "Info",
    NotificationSeverity.SUCCESS: "Success",
    NotificationSeverity.WARNING: "Warning",
    NotificationSeverity.ERROR: "Error",
    NotificationSeverity.CRITICAL: "Critical",
}

ActionSequence = Sequence[Tuple[str, Callable[[], None]]]


logger_call = {
    NotificationSeverity.NONE: logger.debug,
    NotificationSeverity.DEBUG: logger.debug,
    NotificationSeverity.INFO: logger.info,
    NotificationSeverity.SUCCESS: logger.info,
    NotificationSeverity.WARNING: logger.warning,
    NotificationSeverity.ERROR: logger.error,
    NotificationSeverity.CRITICAL: logger.critical,
}


class Notification(Event):
    """A Notification event.  Usually created by :class:`NotificationManager`.

    Parameters
    ----------
    message : str
        The main message/payload of the notification.
    severity : str or NotificationSeverity, optional
        The severity of the notification, by default
        `NotificationSeverity.WARNING`.
    actions : sequence of tuple, optional
        Where each tuple is a `(str, callable)` 2-tuple where the first item
        is a name for the action (which may, for example, be put on a button),
        and the callable is a callback to perform when the action is triggered.
        (for example, one might show a traceback dialog). by default ()
    """

    def __init__(
        self,
        message: str,
        severity: Union[str, NotificationSeverity] = NotificationSeverity.WARNING,
        actions: ActionSequence = (),
        auto_close: bool = True,
        **kwargs,
    ):
        self.severity = NotificationSeverity(severity)
        super().__init__(type=str(self.severity).lower(), **kwargs)
        self.message = message
        self.actions = actions
        self.auto_close = auto_close

        # let's store when the object was created;
        self.date = datetime.now()

    @classmethod
    def from_exception(cls, exc: BaseException, **kwargs) -> Notification:
        """Create notification from error."""
        return ErrorNotification(exc, **kwargs)

    @classmethod
    def from_warning(cls, warning: Warning, **kwargs) -> Notification:
        """Create notification from warning."""
        return WarningNotification(warning, **kwargs)

    def __str__(self):
        return f"{self.date}: {self.message}"

    def as_plain_str(self) -> str:
        """Render as string."""
        return f"{self.date}: {self.message}"


class ErrorNotification(Notification):
    """Error notification."""

    exception: BaseException
    _traceback = None

    def __init__(self, exception: BaseException, *args, **kwargs):
        msg = getattr(exception, "message", str(exception))
        actions = getattr(exception, "actions", ())
        super().__init__(msg, NotificationSeverity.ERROR, actions)
        # extract exception from the tuple
        if isinstance(exception, tuple):
            for value in exception:
                if isinstance(value, Exception):
                    exception = value
                    break
        self.exception = exception
        if hasattr(exception, "__traceback__"):
            self._traceback = exception.__traceback__

    @property
    def traceback(self):
        """Retrieve traceback."""
        if self._traceback is None:
            self._traceback = self.exception.__traceback__
        return self._traceback

    def as_html(self):
        """Render as html."""
        from napari.utils._tracebacks import get_tb_formatter

        fmt = get_tb_formatter()
        exception = self.exception
        exc_info = (
            exception.__class__,
            exception,
            exception.__traceback__,
        )
        return fmt(exc_info, as_html=True)

    def as_plain_str(self) -> str:
        """Render as string."""
        import traceback

        return "".join(traceback.format_stack())

    def __str__(self):
        """Render as string."""
        from napari.utils._tracebacks import get_tb_formatter

        fmt = get_tb_formatter()
        exception = self.exception
        exc_info = (
            exception.__class__,
            exception,
            exception.__traceback__,
        )
        return fmt(exc_info, as_html=False)


class WarningNotification(Notification):
    """Warning notification."""

    warning: Warning

    def __init__(self, warning: Warning, filename=None, lineno=None, *args, **kwargs):
        msg = getattr(warning, "message", str(warning))
        actions = getattr(warning, "actions", ())
        super().__init__(msg, NotificationSeverity.WARNING, actions)
        self.warning = warning
        self.filename = filename
        self.lineno = lineno

    def __str__(self):
        category = type(self.warning).__name__
        return f"{self.filename}:{self.lineno}: {category}: {self.warning}!"


class NotificationManager:
    """
    A notification manager, to route all notifications through.

    Only one instance is in general available through napari; as we need
    notification to all flow to a single location that is registered with the
    sys.except_hook  and showwarning hook.

    This can and should be used a context manager; the context manager will
    properly re-entered, and install/remove hooks and keep them in a stack to
    restore them.

    While it might seem unnecessary to make it re-entrant; or to make the
    re-entrancy no-op; one need to consider that this could be used inside
    another context manager that modify except_hook and showwarning.

    Currently the original except and show warnings hooks are not called; but
    this could be changed in the future; this poses some questions with the
    re-entrency of the hooks themselves.
    """

    records: List[Notification]
    _instance: Optional[NotificationManager] = None

    def __init__(self) -> None:
        self.records: List[Notification] = []
        self.exit_on_error = False
        self.notification_ready = self.changed = EventEmitter(source=self, event_class=Notification)
        self.records_cleared = EventEmitter(source=self, event_class=Event)
        self._originals_except_hooks = []
        self._original_showwarnings_hooks = []
        self._originals_thread_except_hooks = []

    def __enter__(self):
        self.install_hooks()
        return self

    def __exit__(self, *args, **kwargs):
        self.restore_hooks()

    def clear(self):
        """Remove past notifications from the records list."""
        self.records.clear()
        self.records_cleared(Event("clear"))

    def install_hooks(self):
        """
        Install a `sys.excepthook`, a `showwarning` hook and a
        threading.excepthook to display any message in the UI,
        storing the previous hooks to be restored if necessary.
        """
        from ionglow.config import get_settings

        # if getattr(threading, "excepthook", None):
        #     # TODO: we might want to display the additional thread information
        #     self._originals_thread_except_hooks.append(threading.excepthook)
        #     threading.excepthook = self.receive_thread_error
        # else:
        #     # Patch for Python < 3.8
        #     _setup_thread_excepthook()

        if not get_settings().telemetry.enabled:
            self._originals_except_hooks.append(sys.excepthook)
            self._original_showwarnings_hooks.append(warnings.showwarning)
            sys.excepthook = self.receive_error

    # warnings.showwarning = self.receive_warning

    def restore_hooks(self):
        """Remove hooks installed by `install_hooks` and restore previous hooks."""
        if getattr(threading, "excepthook", None):
            # `threading.excepthook` available only for Python >= 3.8
            if self._originals_thread_except_hooks:
                threading.excepthook = self._originals_thread_except_hooks.pop()

        if self._originals_except_hooks:
            sys.excepthook = self._originals_except_hooks.pop()
        if self._original_showwarnings_hooks:
            warnings.showwarning = self._original_showwarnings_hooks.pop()

    def dispatch(self, notification: Notification):
        """Dispatch notification."""
        self.records.append(notification)
        self.notification_ready(notification)

    #         if isinstance(notification, ErrorNotification):
    #             logger_call[notification.severity](notification.as_plain_str())

    def receive_thread_error(self, args: threading.ExceptHookArgs):
        """Receive error from thread."""
        self.receive_error(*args)

    def receive_error(
        self,
        exctype: Optional[Type[BaseException]] = None,
        value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
        thread: Optional[threading.Thread] = None,
    ):
        """Receive error."""
        if isinstance(value, KeyboardInterrupt):
            sys.exit("Closed by KeyboardInterrupt")
        if self.exit_on_error:
            sys.__excepthook__(exctype, value, traceback)
            sys.exit("Exit on error")

        try:
            notif = Notification.from_exception(value)
            self.dispatch(notif)
        except Exception:
            logger.error("Encountered a problem while parsing an error.")

    def receive_warning(
        self,
        message: Warning,
        category: Type[Warning],
        filename: str,
        lineno: int,
        file=None,
        line=None,
    ):
        """Receive warning."""
        self.dispatch(Notification.from_warning(message, filename=filename, lineno=lineno))

    def receive_info(self, message: str):
        """Receive information message."""
        self.dispatch(Notification(message, NotificationSeverity.INFO))


NOTIFICATION_MANAGER: NotificationManager = NotificationManager()


def show_info(message: str):
    """Show message."""
    NOTIFICATION_MANAGER.receive_info(message)


def _setup_thread_excepthook():
    """Workaround for `sys.excepthook` thread bug from: http://bugs.python.org/issue1230540."""
    _init = threading.Thread.__init__

    def init(self, *args, **kwargs):
        _init(self, *args, **kwargs)
        _run = self.run

        def run_with_except_hook(*args2, **kwargs2):
            try:
                _run(*args2, **kwargs2)
            except Exception:
                sys.excepthook(*sys.exc_info())

        self.run = run_with_except_hook

    threading.Thread.__init__ = init



import typing as ty

from qtpy import QtCore, QtGui
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt, Signal


class QtClickLabel(QtW.QLabel):
    """Clickable label without highlight."""

    evt_clicked = Signal()

    def mouseReleaseEvent(self, ev: QtGui.QMouseEvent) -> None:  # type: ignore[override]
        """Emit the clicked signal when the left mouse button is released."""
        if ev.button() == Qt.MouseButton.LeftButton:
            self.evt_clicked.emit()
        return super().mouseReleaseEvent(ev)


class QtVerticalLabel(QtW.QLabel):
    """Rotated label."""

    def __init__(self, *args):
        QtW.QLabel.__init__(self, *args)

    def paintEvent(self, event) -> None:
        painter = QtGui.QPainter(self)
        painter.translate(0, self.height())
        painter.rotate(-90)
        # calculate the size of the font
        fm = QtGui.QFontMetrics(painter.font())
        xoffset = int(fm.boundingRect(self.text()).width() / 2)
        yoffset = int(fm.boundingRect(self.text()).height() / 2)
        x = int(self.width() / 2) + yoffset
        y = int(self.height() / 2) - xoffset
        # because we rotated the label, x affects the vertical placement, and y affects the horizontal
        painter.drawText(y, x, self.text())
        painter.end()

    def minimumSizeHint(self):
        size = super().minimumSizeHint()
        return QtCore.QSize(size.height(), size.width())

    def sizeHint(self):
        size = super().sizeHint()
        return QtCore.QSize(size.height(), size.width())


class QtClickableLabel(QtW.QLabel):
    """A label widget that behaves like a button."""

    evt_clicked = Signal()

    def __init__(self, text: str, parent=None):
        super().__init__(parent)
        self._text = ""
        self._elide_mode = Qt.TextElideMode.ElideNone
        # self.setFixedHeight(24)

        self.setSizePolicy(QtW.QSizePolicy.Policy.Minimum, QtW.QSizePolicy.Policy.Expanding)
        self.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.setText(text)

    def elideMode(self) -> Qt.TextElideMode:
        """The current Qt.TextElideMode."""
        return self._elide_mode

    def setElideMode(self, mode: Qt.TextElideMode) -> None:
        """Set the elide mode to a Qt.TextElideMode."""
        self._elide_mode = Qt.TextElideMode(mode)
        super().setText(self._elidedText())

    def mouseReleaseEvent(self, ev: QtGui.QMouseEvent) -> None:  # type: ignore[override]
        """Emit the clicked signal when the left mouse button is released."""
        if ev.button() == Qt.MouseButton.LeftButton:
            self.evt_clicked.emit()
        return super().mouseReleaseEvent(ev)

    def enterEvent(self, a0: QtCore.QEvent) -> None:
        """Add an underline to the text and change the cursor to a hand."""
        font = self.font()
        font.setUnderline(True)
        self.setFont(font)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.update()
        return super().enterEvent(a0)

    def leaveEvent(self, a0: QtCore.QEvent) -> None:
        """Reset the text and cursor to their original state."""
        font = self.font()
        font.setUnderline(False)
        self.setFont(font)
        self.unsetCursor()
        return super().leaveEvent(a0)

    # Reimplemented QT methods

    def text(self) -> str:
        """Return the label's text.

        If no text has been set this will return an empty string.
        """
        return self._text

    def setText(self, text: str):
        """Set the label's text.

        Setting the text clears any previous content.
        NOTE: we set the QLabel private text to the elided version
        """
        self._text = text
        super().setText(self._elidedText())

    def resizeEvent(self, ev: QtGui.QResizeEvent) -> None:
        """Resize event."""
        ev.accept()
        super().setText(self._elidedText())

    def _elidedText(self) -> str:
        """Return `self._text` elided to `width`."""
        fm = QtGui.QFontMetrics(self.font())
        # the 2 is a magic number that prevents the ellipses from going missing
        # in certain cases (?)
        width = self.width() - 2
        if not self.wordWrap():
            return fm.elidedText(self._text, self._elide_mode, width)

        # get number of lines we can fit without eliding
        nlines = self.height() // fm.height() - 1
        # get the last line (elided)
        text = self._wrappedText()
        last_line = fm.elidedText("".join(text[nlines:]), self._elide_mode, width)
        # join them
        return "".join(text[:nlines] + [last_line])

    def _wrappedText(self) -> ty.List[str]:
        return QtClickableLabel.wrapText(self._text, self.width(), self.font())


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    def _test():
        print("clicked")

    app, frame, ha = qframe(False)
    frame.setMinimumSize(600, 600)
    widget = QtClickableLabel("Test that will be underlined when hovered over", frame)
    widget.evt_clicked.connect(_test)
    ha.addWidget(widget, stretch=True)

    frame.show()
    sys.exit(app.exec_())



"""Mixin class for QtAwesome widgets."""

from __future__ import annotations

import typing as ty

import qtawesome
from loguru import logger
from qtpy.QtCore import QSize

from qtextra.assets import MISSING, get_icon
from qtextra.config import THEMES


class QtaMixin:
    """Mixin class for Qta widgets."""

    _qta_data: tuple | None = None
    _checked_qta_data: tuple | None = None
    _icon_color: str | None = None

    setIcon: ty.Callable
    setMinimumSize: ty.Callable
    setMaximumSize: ty.Callable
    setIconSize: ty.Callable
    setObjectName: ty.Callable

    def _set_icon(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        """Set icon."""
        try:
            icon = qtawesome.icon(*args, **kwargs)
            self.setIcon(icon)
        except Exception as exc:
            logger.warning(f"Failed to set icon: args={args};  kws={kwargs}\n{exc}")
            icon, _ = get_icon(MISSING)  # type: ignore[misc]
            icon = qtawesome.icon(icon, color=THEMES.get_hex_color("warning"))
            self.setIcon(icon)

    def set_qta(self, name: str | tuple[str, dict], **kwargs: ty.Any) -> None:
        """Set QtAwesome icon."""
        name, kwargs_ = get_icon(name)  # type: ignore[misc]
        kwargs.update(kwargs_)
        self._qta_data = (name, kwargs)
        color_ = kwargs.pop("color", None)
        if color_:
            self._icon_color = color_
        color = color_ or self._icon_color or THEMES.get_hex_color("icon")
        if "spin" in kwargs:
            kwargs["animation"] = qtawesome.Spin(self, autostart=True)
            kwargs.pop("spin")
        if "pulse" in kwargs:
            kwargs["animation"] = qtawesome.Pulse(self, autostart=True)
            kwargs.pop("pulse")
        self._set_icon(name, **self._qta_data[1], color=color)

    def _set_qta_icon(self, name: str, **kwargs: ty.Any) -> None:
        """Update icon without setting any attributes."""
        color = self._icon_color or THEMES.get_hex_color("icon")
        self._set_icon(name, **kwargs, color=color)

    def set_default_size(
        self,
        xxsmall: bool = False,
        xsmall: bool = False,
        small: bool = False,
        normal: bool = False,
        average: bool = False,
        medium: bool = False,
        large: bool = False,
        xlarge: bool = False,
        xxlarge: bool = False,
    ) -> None:
        """Set size of the icon."""
        if xxsmall:
            self.set_xxsmall()
        elif xsmall:
            self.set_xsmall()
        elif small:
            self.set_small()
        elif normal:
            self.set_normal()
        elif average:
            self.set_average()
        elif medium:
            self.set_medium()
        elif large:
            self.set_large()
        elif xlarge:
            self.set_xlarge()
        elif xxlarge:
            self.set_xxlarge()

    def set_qta_size(self, size: ty.Tuple[int, int]) -> None:
        """Set maximum size of the icon."""
        size = QSize(*size)  # type: ignore[assignment]
        self.setMinimumSize(size)
        self.setMaximumSize(size)
        self.setIconSize(size)
        self.setObjectName("")

    @classmethod
    def get_icon_size_for_name(cls, name: str) -> tuple[str, tuple[int, int]]:
        """Get icon size for name."""
        if name == "xxsmall":
            return "xxsmall_icon", (10, 10)
        elif name in ("xsmall", "small"):
            return "xsmall_icon", (16, 16)
        elif name == "normal":
            return "normal_icon", (20, 20)
        elif name == "average":
            return "average_icon", (24, 24)
        elif name == "medium":
            return "medium_icon", (28, 28)
        elif name == "large":
            return "large_icon", (32, 32)
        elif name == "xlarge":
            return "xlarge_icon", (60, 60)
        elif name == "xxlarge":
            return "xxlarge_icon", (80, 80)
        elif name == "xxxlarge":
            return "xxxlarge_icon", (120, 120)
        return "average_icon", (24, 24)

    def set_xxsmall(self) -> None:
        """Set large."""
        self.setObjectName("xxsmall_icon")
        self.setIconSize(QSize(10, 10))

    def set_xsmall(self) -> None:
        """Set large."""
        self.setObjectName("xsmall_icon")
        self.setIconSize(QSize(16, 16))

    def set_small(self) -> None:
        """Set large font."""
        self.setObjectName("small_icon")
        self.setIconSize(QSize(16, 16))

    def set_normal(self) -> None:
        """Set medium font."""
        self.setObjectName("normal_icon")
        self.setIconSize(QSize(20, 20))

    def set_average(self) -> None:
        """Set medium font."""
        self.setObjectName("average_icon")
        self.setIconSize(QSize(24, 24))

    def set_medium(self) -> None:
        """Set medium font."""
        self.setObjectName("medium_icon")
        self.setIconSize(QSize(28, 28))

    def set_large(self) -> None:
        """Set large font."""
        self.setObjectName("large_icon")
        self.setIconSize(QSize(32, 32))

    def set_xlarge(self) -> None:
        """Set large."""
        self.setObjectName("xlarge_icon")
        self.setIconSize(QSize(60, 60))

    def set_xxlarge(self) -> None:
        """Set large."""
        self.setObjectName("xxlarge_icon")
        self.setIconSize(QSize(80, 80))

    def set_xxxlarge(self) -> None:
        """Set large."""
        self.setObjectName("xxxlarge_icon")
        self.setIconSize(QSize(120, 120))

    def _update_qta(self) -> None:
        """Update qta icon."""
        if self._qta_data:
            name, kwargs = self._qta_data
            size = self.minimumSize()
            self.set_qta(name, **kwargs)
            self.setMinimumSize(size)

    def _update_from_event(self, event):
        """Update theme based on event."""
        if event.type == "icon":
            self._update_qta()

    # Alias methods to offer Qt-like interface
    _setIcon = _set_icon
    setQta = set_qta
    _setQtaIcon = _set_qta_icon
    getIconSizeForName = get_icon_size_for_name
    setDefaultSize = set_default_size
    setQtaSize = set_qta_size
    setXXSmall = set_xxsmall
    setXSmall = set_xsmall
    setSmall = set_small
    setNormal = set_normal
    setAverage = set_average
    setMedium = set_medium
    setLarge = set_large
    setXLarge = set_xlarge
    setXXLarge = set_xxlarge
    setXXXLarge = set_xxxlarge



"""Searchable combobox."""

from __future__ import annotations

import typing as ty

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QComboBox, QCompleter, QStyledItemDelegate, QWidget


class QtSearchableComboBox(QComboBox):
    """Searchable combobox."""

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.setEditable(True)
        self.completer_object = QCompleter()
        self.completer_object.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.completer_object.setModelSorting(QCompleter.ModelSorting.CaseSensitivelySortedModel)
        self.completer_object.setFilterMode(Qt.MatchFlag.MatchContains)
        self.completer_object.activated.connect(self.on_activated)
        self.setCompleter(self.completer_object)
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)  # ensures that incorrect values are not added
        self.completer_object.popup().setItemDelegate(QStyledItemDelegate(self))
        self.completer_object.popup().setObjectName("search_box_popup")

    def _text_activated(self):  # pragma: no cover
        self.textActivated.emit(self.currentText())

    def on_activated(self, value: str):
        """On activated."""
        self.currentIndexChanged.emit(self.currentIndex())
        self.currentTextChanged.emit(self.currentText())

    def addItem(self, *args):
        """Add item."""
        super().addItem(*args)
        self.completer_object.setModel(self.model())

    def addItems(self, items: ty.Sequence[str]):
        """Add items."""
        super().addItems(items)
        self.completer_object.setModel(self.model())

    def removeItem(self, index: int):
        """Remove item."""
        super().removeItem(index)
        self.completer_object.setModel(self.model())

    # Alias methods to offer Qt-like interface
    _textActivated = _text_activated
    onActivated = on_activated


def add_search_to_combobox(combo: QComboBox) -> None:
    """Add search to combobox."""
    combo.setEditable(True)
    completer_object = QCompleter()
    completer_object.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
    completer_object.setModelSorting(QCompleter.ModelSorting.CaseSensitivelySortedModel)
    completer_object.setFilterMode(Qt.MatchFlag.MatchContains)
    combo.setCompleter(completer_object)
    combo.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)  # ensures that incorrect values are not added
    completer_object.popup().setItemDelegate(QStyledItemDelegate(combo))
    completer_object.popup().setObjectName("search_box_popup")
    combo.completer_object = completer_object
    combo._text_activated = lambda: combo.textActivated.emit(combo.currentText())



"""Popup window for Queue module."""

from __future__ import annotations

from qtpy.QtWidgets import QFormLayout, QWidget

import qtextra.helpers as hp
from qtextra.queue.queue_widget import QUEUE, QueueList
from qtextra.widgets.qt_dialog import QtFramelessTool


class QueuePopup(QtFramelessTool):
    """Popup queue."""

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent=parent)
        self.setMinimumSize(700, 600)

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QFormLayout:
        """Make panel."""
        self.queue_list = QueueList(self)
        self.queue_list.AUTO_EXPAND = True

        self.clear_btn = hp.make_btn(self, "Clear", func=self.queue_list.on_clear_queue, tooltip="Clear all tasks")
        self.n_tasks = hp.make_labelled_slider(
            self,
            minimum=1,
            maximum=max(QUEUE.n_parallel, 6),
            value=QUEUE.n_parallel,
            tooltip="Maximum number of tasks to run simultaneously.",
            func=QUEUE.set_max_parallel,
        )

        layout = hp.make_form_layout()
        layout.setContentsMargins(6, 6, 6, 6)
        layout.addRow(self._make_hide_handle("Task queue")[1])
        layout.addRow("Number of parallel tasks", self.n_tasks)
        layout.addRow(self.clear_btn)
        layout.addRow(hp.make_h_line(self))
        layout.addRow(self.queue_list)
        return layout


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.config import THEMES
        from qtextra.queue.task import Task
        from qtextra.utils.dev import qapplication

        _ = qapplication()  # analysis:ignore
        dlg = QueuePopup(None)
        THEMES.set_theme_stylesheet(dlg)

        for i in range(3):
            task = Task(f"Task {i}", [["echo", "Task", f"{i}"], ["sleep", "3"], ["sleep", "3"], ["sleep", "3"]])
            QUEUE.add_task(task)

        dlg.show()
        sys.exit(dlg.exec_())

    _main()  # type: ignore[no-untyped-call]



"""Assets."""

from __future__ import annotations

import typing as ty
from pathlib import Path

from koyo.utilities import get_module_path
from loguru import logger

HERE = Path(get_module_path("qtextra.assets", "__init__.py")).parent.resolve()

ICON_PATH = HERE / "icons"
ICON_PATH.mkdir(exist_ok=True)
ICONS = {x.stem: str(x) for x in ICON_PATH.iterdir() if x.suffix == ".svg"}

STYLE_PATH = HERE / "stylesheets"
STYLE_PATH.mkdir(exist_ok=True)
STYLES = {x.stem: str(x) for x in STYLE_PATH.iterdir() if x.suffix == ".qss"}

# Some gifs were made using https://loading.io/
# orange color = #ff4500
LOADING_SQUARE_GIF = str(HERE / "gifs" / "loading-square.gif")
LOADING_CIRCLE_GIF = str(HERE / "gifs" / "loading-circle.gif")
LOADING_DOTS_GIF = str(HERE / "gifs" / "loading-dots.gif")
LOADING_INFINITY_GIF = str(HERE / "gifs" / "loading-infinity.gif")
LOADING_OVAL_GIF = str(HERE / "gifs" / "loading-oval.gif")

LOADING_GIFS = {
    "square": LOADING_SQUARE_GIF,
    "circle": LOADING_CIRCLE_GIF,
    "dots": LOADING_DOTS_GIF,
    "infinity": LOADING_INFINITY_GIF,
    "oval": LOADING_OVAL_GIF,
}

MISSING = "MISSING"

QTA_MAPPING: ty.Dict[str, str | tuple[str, dict]] = {
    MISSING: "ri.error-warning-line",
    "clear": "mdi6.delete-empty",
    "json": "mdi6.code-json",
    "binary": "msc.file-binary",
    "hdf5": "fa5s.file",
    "open": "fa5s.folder-open",
    "folder": "mdi.folder-move-outline",
    "cross": "fa5s.times",
    "cross_full": "fa5s.times-circle",
    "minimise": "fa5s.window-minimize",
    "help": "mdi.help-circle-outline",
    "clipboard": "fa5s.clipboard-list",
    "colorbar": "mdi.invert-colors",
    "gear": "ph.gear-fill",
    "zoom_out": "fa5s.expand",
    "ruler": "ph.ruler",
    "ipython": "mdi.console",
    "cli": "mdi6.console-line",
    "template": "ei.file-new",
    # "log": "mdi6.file-document",
    "log": "mdi6.clipboard-text",
    "filter_and": "mdi6.filter-plus",
    "filter_or": "mdi6.filter-minus",
    "and": "mdi6.ampersand",
    "or": "fa5s.grip-lines-vertical",
    # "ruler": "fa5s.ruler-horizontal",
    "text": "mdi.format-text",
    "crosshair": "ph.crosshair-simple",
    "none": "mdi6.cancel",
    "move": "ei.move",
    "move_handle": "ei.move",
    "lasso": "mdi6.lasso",
    "marker": "fa5s.map-marker-alt",
    "zoom": "mdi.magnify",
    "erase": "ph.eraser-fill",
    "new": "mdi.new-box",
    # "check": "fa5s.check",
    "edit": "ri.edit-box-fill",
    "add": "ri.add-circle-line",
    "remove": "ri.indeterminate-circle-line",
    "check": "ri.checkbox-circle-line",
    "delete": "fa5s.trash-alt",
    "shuffle": "ph.shuffle-bold",
    "picker": "mdi6.eyedropper",
    "paint": "fa5s.paint-brush",
    "fill": "fa5s.fill-drip",
    "cancel": "mdi.close-circle",
    "paint_palette": "ph.palette-fill",
    "copy_to_clipboard": "mdi.clipboard-arrow-left-outline",
    "copy": "fa6.copy",
    "copy_all": "fa6s.copy",
    "sort": "fa5s.sort-amount-up",
    "sort_ascending": "fa5s.sort-amount-up",
    "sort_descending": "fa5s.sort-amount-down",
    "graph": "mdi.graph",
    "reload": "mdi6.cached",
    "replace": "msc.replace",
    "save": "fa5s.save",
    "screenshot": "mdi.camera-outline",
    "github": "fa5b.github",
    "request": "msc.request-changes",
    "web": "mdi.web",
    "bug": "fa5s.bug",
    "info": "fa5s.info-circle",
    "warning": "fa5s.exclamation-triangle",
    "error": "fa5s.times-circle",
    "critical": "fa5s.times-circle",
    "debug": "ph.megaphone",
    "success": "fa5s.check",
    "true": "mdi6.check-circle-outline",
    "false": "mdi6.close-circle-outline",
    "axes_label": "mdi.axis-arrow",
    "close": "fa5s.trash-alt",
    "list": "mdi6.format-list-bulleted",
    "pin": "fa5s.map-pin",
    "repeat": "fa5s.redo",
    "more": "mdi.dots-horizontal-circle",
    "previous": "fa5s.chevron-circle-left",
    "next": "fa5s.chevron-circle-right",
    "star": "fa5s.star",
    "cpu": "ri.cpu-line",
    "ram": "fa5s.memory",
    "upgrade": "ei.download",
    "schema": "mdi6.badge-account-horizontal",
    "locked": "mdi.lock",
    "unlocked": "mdi.lock-open",
    "notified": "mdi6.bell",
    "not_notified": "mdi6.bell-off",
    # notification
    "notification": "mdi6.bell",
    "notification_on": "mdi6.bell",
    "notification_off": "mdi6.bell-badge",
    "notification_dismiss": "mdi6.bell-cancel",
    "notification_check": "mdi6.bell-check",
    # translate
    "rotate_left": "fa6s.rotate-left",
    "rotate_right": "fa6s.rotate-right",
    "translate_left": "fa5s.arrow-left",
    "translate_right": "fa5s.arrow-right",
    "translate_up": "fa5s.arrow-up",
    "translate_down": "fa5s.arrow-down",
    "flip_lr": "fa5s.arrows-alt-h",
    "flip_ud": "fa5s.arrows-alt-v",
    "group": "mdi6.group",
    "ungroup": "mdi6.ungroup",
    # app
    "settings": "mdi6.tools",
    "reset": "mdi.lock-reset",
    # "update": "ei.refresh",
    "update": "fa5s.redo-alt",
    "version": "mdi6.update",
    "telemetry": "mdi.telegram",
    "shortcut": "mdi6.tooltip-text",
    "feedback": "msc.feedback",
    "handshake": "fa5.handshake",
    "dev": "mdi6.code-braces",
    "import": "mdi.file-import",
    "export": "mdi.file-export",
    # arrows
    "arrow_up": "fa5s.arrow-up",
    "arrow_down": "fa5s.arrow-down",
    "arrow_left": "fa5s.arrow-left",
    "arrow_right": "fa5s.arrow-right",
    "long_arrow_up": "fa5s.long-arrow-alt-up",
    "long_arrow_down": "fa5s.long-arrow-alt-down",
    "long_arrow_left": "fa5s.long-arrow-alt-left",
    "long_arrow_right": "fa5s.long-arrow-alt-right",
    "caret_left": "fa5s.caret-left",
    "caret_right": "fa5s.caret-right",
    "caret_up": "fa5s.caret-up",
    "caret_down": "fa5s.caret-down",
    # run/play
    "run": "mdi6.run-fast",
    "start": "fa5s.play-circle",
    "retry": "fa5s.redo-alt",
    "stop": "fa5s.stop-circle",
    "pause": "fa5s.pause-circle",
    # "queue": "ph.queue-fill",
    "queue": "mdi6.human-queue",
    # lock
    "lock_closed": "fa5s.lock",
    "lock_open": "fa5s.lock-open",
    # theme
    "light_theme": "ri.sun-fill",
    "dark_theme": "ri.moon-clear-fill",
    # chevrons
    "chevron_down": "fa5s.chevron-down",
    "chevron_up": "fa5s.chevron-up",
    "chevron_left": "fa5s.chevron-left",
    "chevron_right": "fa5s.chevron-right",
    "chevron_down_circle": "fa5s.chevron-circle-down",
    "chevron_up_circle": "fa5s.chevron-circle-up",
    "chevron_left_circle": "fa5s.chevron-circle-left",
    "chevron_right_circle": "fa5s.chevron-circle-right",
    # toggles
    "toggle_on": "mdi.checkbox-marked",
    "toggle_off": "mdi.checkbox-blank-outline",
    # visible
    "visible": "ei.eye-open",
    "visible_on": "ei.eye-open",
    "visible_off": "ei.eye-close",
    # priority
    "priority_high": "mdi6.chevron-triple-up",
    "priority_normal": "ph.equals-fill",
    "priority_low": "mdi6.chevron-triple-down",
    # viewer
    "new_line": "msc.pulse",
    "new_surface": "ei.star",
    "new_labels": "fa5s.tag",
    "new_points": "mdi.scatter-plot",
    "new_shapes": "fa5s.shapes",
    "new_inf_line": "mdi.infinity",
    "new_centroids": "ri.bar-chart-fill",
    "new_region": "ri.bar-chart-horizontal-fill",
    "ndisplay_off": "ph.square",
    "ndisplay_on": "ph.cube",
    "roll": "mdi6.rotate-right-variant",
    "transpose": "ri.t-box-line",
    "grid_off": "mdi6.grid-off",
    "grid_on": "mdi6.grid",
    "home": "fa5s.home",
    "pan_zoom": "ei.move",
    "select": "fa5s.location-arrow",
    "add_points": "ri.add-circle-fill",
    "select_points": "fa5s.location-arrow",
    "delete_shape": "fa5s.times",
    "move_back": "mdi6.arrange-send-backward",
    "move_front": "mdi6.arrange-bring-to-front",
    "line": "ph.line-segment-fill",
    "path": "mdi.chart-line-variant",
    "vertex_insert": "mdi.map-marker-plus",
    "vertex_remove": "mdi.map-marker-minus",
    "vertex_select": "mdi.map-marker-check",
    "grid": "mdi.grid",
    "layers": "fa5s.layer-group",
    "rectangle": "ph.rectangle-bold",
    "ellipse": "mdi6.ellipse-outline",
    "polygon": "mdi.pentagon-outline",
    # selection
    "invert_selection": "fa5s.exchange-alt",
    "pin_on": ("ph.push-pin-fill", {"rotated": -45}),
    "pin_off": "ph.push-pin-fill",
    "minimize": "fa5s.window-minimize",
    "maximize": "fa5s.window-maximize",
    "fullscreen": "fa5s.expand",
    "random": "fa5s.random",
    "color_palette": "fa5s.palette",
    "magic": "mdi6.auto-fix",
    "approve": "fa5s.thumbs-up",
    "reject": "fa5s.thumbs-down",
    "target": "mdi.target",
}


def update_icon_mapping(mapping: ty.Dict[str, str]) -> None:
    """Update icon mapping."""
    global QTA_MAPPING
    for k, v_new in mapping.items():
        v_exist = QTA_MAPPING.get(k, None)
        v_new = mapping[k]
        if v_exist and v_exist == v_new:
            print(f"Warning: Icon mapping already exists for '{k}'")
        QTA_MAPPING[k] = v_new


def update_styles(mapping: ty.Dict[str, str], append: bool = True) -> None:
    """Update icon mapping."""
    global STYLES

    if append:
        STYLES.update(mapping)
    else:
        mapping.update(STYLES)
        STYLES = mapping


def update_icons(mapping: ty.Dict[str, str], append: bool = True) -> None:
    """Update icon mapping."""
    global ICONS

    if append:
        ICONS.update(mapping)
    else:
        mapping.update(ICONS)
        ICONS = mapping


def get_icon(name: str | tuple[str, dict]) -> tuple[str, dict]:
    """Return icon."""
    kwargs = None
    if isinstance(name, tuple):
        name, kwargs = name
    if kwargs is None:
        kwargs = {}

    original_name = name
    if name == "":
        name = QTA_MAPPING[MISSING]
    elif "." not in name:
        name = QTA_MAPPING.get(name)
        if name is None:
            logger.warning(f"Failed to retrieve icon: '{original_name}'")
            name = QTA_MAPPING[MISSING]
    if isinstance(name, tuple):
        name, kwargs_ = name
        kwargs.update(kwargs_)
    return name, kwargs


def get_stylesheet(theme: ty.Optional[str] = None, extra: ty.Optional[ty.List[str]] = None) -> str:
    """Combine all qss files into single, possibly pre-themed, style string.

    Parameters
    ----------
    theme : str, optional
        Theme to apply to the stylesheet. If no theme is provided, the returned
        stylesheet will still have ``{{ template_variables }}`` that need to be
        replaced using the :func:`qtextra.template` function prior
        to using the stylesheet.
    extra : list of str, optional
        Additional paths to QSS files to include in stylesheet, by default None

    Returns
    -------
    css : str
        The combined stylesheet.
    """
    stylesheet = ""
    try:
        for key in sorted(STYLES):
            file = STYLES[key]
            with open(file) as f:
                stylesheet += f.read()
        if extra:
            for file in extra:
                with open(file) as f:
                    stylesheet += f.read()
    except FileNotFoundError as e:
        logger.error(f"Failed to load stylesheet: {e}")

    if theme:
        from qtextra.config.theme import THEMES
        from qtextra.utils.template import template

        return template(stylesheet, **THEMES.get_theme(theme, as_dict=True))
    return stylesheet



"""Info widget."""

from qtpy.QtWidgets import QFormLayout, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_dialog import QtFramelessPopup


class InfoDialog(QtFramelessPopup):
    """Dialog to display some useful information."""

    def __init__(self, parent: QWidget, text: str):
        super().__init__(parent)
        self.setMinimumWidth(300)
        self.setMinimumHeight(200)
        self.label.setText(text)

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QFormLayout:
        """Make panel."""
        header_layout = self._make_title_handle("Help information")

        self.label = hp.make_scrollable_label(
            self, wrap=True, enable_url=True, selectable=True, object_name="title_label"
        )

        layout = hp.make_form_layout(parent=self)
        layout.addRow(header_layout)
        layout.addRow(self.label)
        return layout



"""Event loop."""

import os
import sys
from typing import Optional
from warnings import warn

from qtpy.QtCore import Qt
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QApplication

from qtextra import __version__

APP_ID: str = os.getenv("QTEXTRA_APP_ID", f"qtextra.qtextra.{__version__}")
APP_NAME: str = os.getenv("QTEXTRA_APP_NAME", "qtextra")
APP_ORG_NAME: str = os.getenv("QTEXTRA_APP_ORG_NAME", "qtextra")
APP_ORG_DOMAIN: str = os.getenv("QTEXTRA_APP_ORG_DOMAIN", "")


# store reference to QApplication to prevent garbage collection
_app_ref = None


def set_app_id(app_id):
    """Set app ID."""
    if os.name == "nt" and app_id and not getattr(sys, "frozen", False):
        import ctypes

        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(app_id)


_defaults = {
    "app_name": APP_NAME,
    "app_version": __version__,
    # "icon": ICON_PNG,
    "org_name": APP_ORG_NAME,
    "org_domain": APP_ORG_DOMAIN,
    "app_id": APP_ID,
}


def get_app(
    *,
    app_name: Optional[str] = None,
    app_version: Optional[str] = None,
    icon: Optional[str] = None,
    org_name: Optional[str] = None,
    org_domain: Optional[str] = None,
    app_id: Optional[str] = None,
) -> QApplication:
    """Get or create the Qt QApplication.

    There is only one global QApplication instance, which can be retrieved by
    calling get_app again, (or by using QApplication.instance())

    Parameters
    ----------
    app_name : str, optional
        Set app name (if creating for the first time), by default 'qtextra'
    app_version : str, optional
        Set app version (if creating for the first time), by default __version__
    icon : str, optional
        Set app icon (if creating for the first time), by default
        ICON_PNG
    org_name : str, optional
        Set organization name (if creating for the first time), by default
        'qtextra'
    org_domain : str, optional
        Set organization domain (if creating for the first time), by default
        'qtextra.org'
    app_id : str, optional
        Set organization domain (if creating for the first time).  Will be
        passed to set_app_id (which may also be called independently), by
        default APP_ID

    Returns
    -------
    QApplication
        [description]

    Notes
    -----
    Substitutes QApplicationWithTracing when the QTEXTRA_PERFMON env variable
    is set.

    If the QApplication already exists, we call convert_app_for_tracing() which
    deletes the QApplication and creates a new one. However here with get_app
    we need to create the correct QApplication up front, or we will crash
    because we'd be deleting the QApplication after we created QWidgets with
    it, such as we do for the splash screen.
    """
    # qtextra defaults are all-or nothing.  If any of the keywords are used
    # then they are all used.
    set_values = {k for k, v in locals().items() if v}
    kwargs = locals() if set_values else _defaults
    global _app_ref

    # Note: this MUST be set before the QApplication is instantiated

    app = QApplication.instance()
    if app:
        set_values.discard("ipy_interactive")
        if set_values:
            warn(
                f"QApplication already existed, these arguments to to 'get_app' were ignored: {set_values}",
                stacklevel=2,
            )
    else:
        # automatically determine monitor DPI.
        # Note: this MUST be set before the QApplication is instantiated
        os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
        if hasattr(Qt, "AA_EnableHighDpiScaling"):
            QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
        if hasattr(Qt, "AA_UseHighDpiPixmaps"):
            QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
        if hasattr(Qt, "AA_UseStyleSheetPropagationInWidgetStyles"):
            QApplication.setAttribute(Qt.AA_UseStyleSheetPropagationInWidgetStyles, True)
        if hasattr(Qt, "AA_ShareOpenGLContexts"):
            QApplication.setAttribute(Qt.AA_ShareOpenGLContexts, True)
        app = QApplication(sys.argv)

    # if this is the first time the Qt app is being instantiated, we set
    # the name and metadata
    app.setApplicationName(kwargs.get("app_name"))
    app.setApplicationVersion(kwargs.get("app_version"))
    app.setOrganizationName(kwargs.get("org_name"))
    app.setOrganizationDomain(kwargs.get("org_domain"))
    app.setWindowIcon(QIcon(kwargs.get("icon")))
    set_app_id(kwargs.get("app_id"))

    # if perf_config and not perf_config.patched:
    #     # Will patch based on config file.
    #     perf_config.patch_callables()

    # if not _ipython_has_eventloop():
    #     NOTIFICATION_MANAGER.notification_ready.connect(QtNotification.show_notification)
    # NOTIFICATION_MANAGER.notification_ready.connect(show_console_notification)

    if app.windowIcon().isNull():
        app.setWindowIcon(QIcon(kwargs.get("icon")))

    if not _app_ref:  # running get_app for the first time
        # When a new theme is added, at it to the search path.
        from qtpy.QtCore import QDir

        from qtextra.config import THEMES

        # see docstring of `wait_for_workers_to_quit` for caveats on killing
        # workers at shutdown.
        # app.aboutToQuit.connect(wait_for_workers_to_quit)
        # Setup search paths for currently installed themes.
        for name in THEMES.themes:
            QDir.addSearchPath(f"theme_{name}", str(THEMES.get_theme_path(name)))

    _app_ref = app  # prevent garbage collection
    return app


def _ipython_has_eventloop() -> bool:
    """Return True if IPython %gui qt is active.

    Using this is better than checking ``QApp.thread().loopLevel() > 0``,
    because IPython starts and stops the event loop continuously to accept code
    at the prompt.  So it will likely "appear" like there is no event loop
    running, but we still don't need to start one.
    """
    try:
        from IPython import get_ipython

        return get_ipython().active_eventloop == "qt"
    except (ImportError, AttributeError):
        return False


def run(*, force=False, max_loop_level=1, _func_name="run"):
    """Start the Qt Event Loop.

    Parameters
    ----------
    force : bool, optional
        Force the application event_loop to start, even if there are no top
        level widgets to show.
    max_loop_level : int, optional
        The maximum allowable "loop level" for the execution thread.  Every
        time `QApplication.exec_()` is called, Qt enters the event loop,
        increments app.thread().loopLevel(), and waits until exit() is called.
        This function will prevent calling `exec_()` if the application already
        has at least ``max_loop_level`` event loops running.  By default, 1.
    _func_name : str, optional
        name of calling function, by default 'run'.  This is only here to
        provide functions like `gui_qt` a way to inject their name into the
        warning message.

    Raises
    ------
    RuntimeError
        (To avoid confusion) if no widgets would be shown upon starting the
        event loop.
    """
    # from ionglow.utils.notifications import NOTIFICATION_MANAGER

    from qtextra.dialogs.sentry import install_error_monitor

    if _ipython_has_eventloop():
        # If %gui qt is active, we don't need to block again.
        return

    app = QApplication.instance()
    if not app:
        raise RuntimeError(
            "No Qt app has been created. One can be created by calling `get_app()` or qtpy.QtWidgets.QApplication([])"
        )
    if not app.topLevelWidgets() and not force:
        warn(
            "Refusing to run a QApplication with no topLevelWidgets. "
            f"To run the app anyway, use `{_func_name}(force=True)`",
            stacklevel=2,
        )
        return

    if app.thread().loopLevel() >= max_loop_level:
        loops = app.thread().loopLevel()
        s = "s" if loops > 1 else ""
        warn(
            f"A QApplication is already running with {loops} event loop{s}."
            "To enter *another* event loop, use "
            f"`{_func_name}(max_loop_level={loops + 1})`",
            stacklevel=2,
        )
        return
    install_error_monitor()
    app.exec_()



"""Directory manager."""

import os
from typing import Dict, List

from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QFrame, QHBoxLayout, QScrollArea, QSizePolicy, QVBoxLayout, QWidget

import qtextra.helpers as hp


class QtDirectoryWidget(QFrame):
    """Directory widget."""

    evt_checked = Signal(str, bool)

    def __init__(self, path: str, exist_obj_name: str = "success", parent=None):
        """Directory widget where the path is displayed alongside a couple of helpful icons/buttons.

        Parameters
        ----------
        path : str
            Path to the directory.
        exist_obj_name : str
            Specifies how the text should be displayed in case the path exists or not. For instance, it can be used to
            highlight when path exists and it should not (e.g. `warning`) or vice-versa (e.g. `success`).
        parent : QWidget, optional
            Specifies the parent of the widget.
        """
        super().__init__(parent=parent)
        self.setMinimumHeight(25)

        self.checkbox = hp.make_checkbox(self, tooltip="Click here to check item")
        self.checkbox.stateChanged.connect(self._on_check)

        self.path_label = hp.make_eliding_label2(
            self, path, elide=Qt.TextElideMode.ElideRight, tooltip="Directory path"
        )
        self.path_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.MinimumExpanding)
        self.path_label.setMinimumWidth(600)

        self.new_icon = hp.make_qta_btn(self, "new", tooltip="This directory does not exist and will be created.")
        self.new_icon.setVisible(False)

        self.warning_icon = hp.make_qta_label(
            self, "warning", tooltip="This directory already exists and it should not."
        )
        self.warning_icon.setVisible(False)

        self.open_btn = hp.make_qta_btn(self, "folder", tooltip="Click here to open directory")
        self.open_btn.clicked.connect(self._on_open)

        self.edit_btn = hp.make_qta_btn(self, "edit", tooltip="Click here to edit path")
        self.edit_btn.clicked.connect(self._on_edit)
        self.edit_btn.setVisible(False)

        self.row = QHBoxLayout()
        self.row.addWidget(self.checkbox, alignment=Qt.AlignmentFlag.AlignVCenter)
        self.row.addWidget(self.path_label, stretch=True, alignment=Qt.AlignmentFlag.AlignLeft)
        self.row.addWidget(self.new_icon)
        self.row.addWidget(self.warning_icon)
        self.row.addWidget(self.open_btn)
        self.row.addWidget(self.edit_btn)
        self.row.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.row)

        self._exist_obj_name = exist_obj_name
        self._not_exist_obj_name = "warning" if exist_obj_name == "success" else "success"
        self.path = path

    @property
    def is_checked(self) -> bool:
        """Flag to indicate if its checked."""
        return self.checkbox.isChecked()

    @property
    def path(self) -> str:
        """Return path."""
        return self._path

    @property
    def is_new(self) -> bool:
        """Check whether specified path will be new."""
        return not os.path.exists(self.path)

    @path.setter
    def path(self, value: str):
        self._path = value
        self.path_label.setText(value)
        self.path_label.setToolTip(self._path)
        exists = os.path.exists(value)
        obj_name = self._exist_obj_name if exists else self._not_exist_obj_name
        self.warning_icon.setVisible(obj_name == "warning")
        self.set_style(obj_name)
        self.open_btn.setVisible(exists)

    def set_style(self, object_name: str):
        """Set new style."""
        hp.update_widget_style(self.path_label, object_name)
        self.warning_icon.setVisible(object_name == "warning")

    def set_new(self, visible: bool):
        """Set visible icon."""
        self.new_icon.setVisible(visible)

    def show_as_path(self, show_full: bool):
        """Show basename without full path."""
        path = self._path if show_full else os.path.basename(self._path)
        self.path_label.setText(path)

    def _on_check(self):
        """Checked/unchecked event."""
        self.evt_checked.emit(self._path, self.checkbox.isChecked())
        hp.disable_with_opacity(self, [self.path_label], not self.checkbox.isChecked())

    def _on_edit(self):
        """Edit value."""
        new_path = hp.get_text(self, "Modify the current path...", "Modify the current path", self.path)
        if new_path is not None:
            self.path = new_path

    def _on_open(self):
        """Open directory."""
        from qtextra.utils.utilities import open_directory

        if os.path.exists(self._path):
            open_directory(self._path)


class QtDirectoryManager(QScrollArea):
    """Directory manager."""

    # triggered whenever new path is added
    evt_added = Signal(str)
    # triggered whenever path is updated
    evt_update = Signal(str)
    # triggered whenever path is removed
    evt_removed = Signal(str)

    def __init__(self, parent=None, exist_obj_name: str = "success"):
        """Directory manager.

        Parameters
        ----------
        parent : QWidget
            parent object
        exist_obj_name : str
            name of the object if the path exists. If value is `success`, then the it will be green and in case it does
            not exist, then the path will be set to `warning` and it will be rendered in red.
        """
        super().__init__(parent=parent)
        self.widgets: Dict[str, QtDirectoryWidget] = {}
        self._exist_obj_name = exist_obj_name

        scroll_widget = QWidget()
        self.setWidget(scroll_widget)

        main_layout = QVBoxLayout(scroll_widget)
        main_layout.setSpacing(2)
        main_layout.addStretch(1)

        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self._layout = main_layout

    @property
    def paths(self) -> List[str]:
        """Return list of paths."""
        return list(self.widgets.keys())

    @property
    def checked_paths(self) -> List[str]:
        """Return list of paths that are checked."""
        return [widget.path for widget in self.widgets.values() if widget.is_checked]

    @property
    def n_new(self) -> int:
        """Return the total count of new directories."""
        return sum([widget.is_new for widget in self.widgets.values()])

    def clear(self):
        """Remove all directories."""
        paths = self.paths
        for path in paths:
            self.remove_path(path)

    def add_path(self, path: str):
        """Add path widget.

        Parameters
        ----------
        path : str
            path to be added to the widget
        """
        widget = QtDirectoryWidget(path, exist_obj_name=self._exist_obj_name, parent=self)
        widget.checkbox.setChecked(True)
        self.widgets[path] = widget
        self._layout.insertWidget(0, widget)
        self.evt_added.emit(path)

    def update_path(self, old_path: str, new_path: str):
        """Update path widget."""
        widget = self.widgets.pop(old_path, None)
        if widget:
            self.widgets[new_path] = widget
            self.widgets.path = new_path
            self.evt_update.emit(new_path)

    def remove_path(self, path: str):
        """Remove widget."""
        widget = self.widgets.pop(path, None)
        if widget:
            self._layout.removeWidget(widget)
            widget.deleteLater()
            self.evt_removed.emit(path)

    def validate(self, path: str, object_name: str):
        """Update style of the directory by setting its object name."""
        widget = self.widgets.get(path, None)
        if widget:
            widget.set_style(object_name)

    def set_new(self, path: str, visible: bool):
        """Update new state of the directory."""
        widget = self.widgets.get(path, None)
        if widget:
            widget.set_new(visible)

    def show_as_path(self, show_full: bool):
        """Show basename of each widget."""
        for widget in self.widgets.values():
            widget.show_as_path(show_full)

    def is_checked(self, path: str) -> bool:
        """Check whether directory is checked."""
        widget = self.widgets.get(path, None)
        if widget:
            return widget.is_checked
        return False


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        def _check():
            print(widget.checked_paths)

        import sys

        from qtextra.helpers import make_btn
        from qtextra.utils.dev import qframe

        app, frame, ha = qframe(False)
        frame.setMinimumHeight(600)
        widget = QtDirectoryManager()
        ha.addWidget(widget)
        for i in range(3):
            widget.add_path(f"PATH {i}")

        btn = make_btn(frame, "Show")
        btn.clicked.connect(_check)
        ha.addWidget(btn)

        frame.show()
        sys.exit(app.exec_())

    _main()



import pytest
from loguru import logger

from qtextra.dialogs.qt_logger import QtLoggerDialog


@pytest.fixture
def qt_logger(qtbot):
    widget = QtLoggerDialog(None)
    qtbot.addWidget(widget)
    return widget


def test_qt_logger_01(qtbot):
    widget = QtLoggerDialog(None)
    qtbot.addWidget(widget)


def test_qt_logger_02(qt_logger, qtbot):
    logger.debug("debug")
    logger.info("INFO")
    logger.warning("warning")
    logger.error("error")
    logger.critical("critical")

    expected = "light"
    qt_logger.logger.set_theme(expected)
    assert qt_logger.logger.THEME == expected

    expected = "dark"
    qt_logger.logger.set_theme(expected)
    assert qt_logger.logger.THEME == expected


def test_qt_logger_03(qt_logger, qtbot):
    expected = "other"
    with pytest.raises(ValueError) as __:
        qt_logger.logger.set_theme(expected)


if __name__ == "__main__":
    pytest.main()



"""Context managers."""

import signal
import socket
from contextlib import contextmanager

from qtpy.QtCore import QSocketNotifier


@contextmanager
def _maybe_allow_interrupt(qapp):
    """
    This manager allows to terminate a plot by sending a SIGINT. It is
    necessary because the running Qt backend prevents Python interpreter to
    run and process signals (i.e., to raise KeyboardInterrupt exception). To
    solve this one needs to somehow wake up the interpreter and make it close
    the plot window. We do this by using the signal.set_wakeup_fd() function
    which organizes a write of the signal number into a socketpair connected
    to the QSocketNotifier (since it is part of the Qt backend, it can react
    to that write event). Afterwards, the Qt handler empties the socketpair
    by a recv() command to re-arm it (we need this if a signal different from
    SIGINT was caught by set_wakeup_fd() and we shall continue waiting). If
    the SIGINT was caught indeed, after exiting the on_signal() function the
    interpreter reacts to the SIGINT according to the handle() function which
    had been set up by a signal.signal() call: it causes the qt_object to
    exit by calling its quit() method. Finally, we call the old SIGINT
    handler with the same arguments that were given to our custom handle()
    handler.
    We do this only if the old handler for SIGINT was not None, which means
    that a non-python handler was installed, i.e. in Julia, and not SIG_IGN
    which means we should ignore the interrupts.

    code from https://github.com/matplotlib/matplotlib/pull/13306
    """
    old_sigint_handler = signal.getsignal(signal.SIGINT)
    handler_args = None
    if old_sigint_handler in (None, signal.SIG_IGN, signal.SIG_DFL):
        yield
        return

    wsock, rsock = socket.socketpair()
    wsock.setblocking(False)
    old_wakeup_fd = signal.set_wakeup_fd(wsock.fileno())
    sn = QSocketNotifier(rsock.fileno(), QSocketNotifier.Type.Read)

    # Clear the socket to re-arm the notifier.
    sn.activated.connect(lambda *args: rsock.recv(1))

    def handle(*args):
        nonlocal handler_args
        handler_args = args
        qapp.exit()

    signal.signal(signal.SIGINT, handle)
    try:
        yield
    finally:
        wsock.close()
        rsock.close()
        sn.setEnabled(False)
        signal.set_wakeup_fd(old_wakeup_fd)
        signal.signal(signal.SIGINT, old_sigint_handler)
        if handler_args is not None:
            old_sigint_handler(*handler_args)



"""Utilities for QtExtra widgets."""

from __future__ import annotations

import os
import sys
import typing as ty

from koyo.utilities import is_installed
from loguru import logger
from qtpy.QtCore import QEvent, Qt, QTimer, Signal
from qtpy.QtWidgets import QApplication, QDialog, QDockWidget, QLayout, QMainWindow, QWidget

if ty.TYPE_CHECKING:
    from qtreload.qt_reload import QtReloadWidget


DEFAULT_MODULES = ("qtextra", "koyo")
if is_installed("qtextraplot"):
    DEFAULT_MODULES += ("qtextraplot",)


def exec_(app: QApplication) -> None:
    from napari._qt.utils import _maybe_allow_interrupt

    with _maybe_allow_interrupt(app):
        sys.exit(app.exec_())


def disable_warnings() -> None:
    """Disable warnings."""
    import warnings

    # disable warnings
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="vispy")
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="pydantic")
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="numpy")
    warnings.filterwarnings("ignore", category=FutureWarning, module="shiboken2")
    warnings.filterwarnings("ignore", category=FutureWarning, module="pandas")
    warnings.filterwarnings("ignore", category=FutureWarning, module="xgboost")
    warnings.filterwarnings("ignore", category=ResourceWarning, module="sentry_sdk")


def qdev(
    parent=None, modules: ty.Iterable[str] = DEFAULT_MODULES, log_func: ty.Callable[[str], None] | None = None
) -> QtReloadWidget:
    """Create reload widget."""
    from qtreload.qt_reload import QtReloadWidget

    from qtextra.config.theme import THEMES

    widget = QtReloadWidget(modules, parent=parent, log_func=log_func)
    widget.evt_stylesheet.connect(THEMES.evt_qss_changed.emit)
    return widget


def qdev_dock(
    parent=None, modules: ty.Iterable[str] = DEFAULT_MODULES, log_func: ty.Callable[[str], None] | None = None
) -> ty.Tuple[QtReloadWidget, QDockWidget]:
    """Create reload widget in dock."""
    widget = qdev(parent, modules, log_func=log_func)
    dock = QDockWidget("Reload", widget)
    return widget, dock


def qdev_popup(
    parent=None, modules: ty.Iterable[str] = DEFAULT_MODULES, log_func: ty.Callable[[str], None] | None = None
) -> ty.Tuple[QtReloadWidget, QWidget]:
    """Create reload widget in popup."""
    widget = qdev(parent, modules, log_func=log_func)
    popup = QDialog()
    popup.setLayout(widget.layout())
    return widget, popup


def qapplication(test_time: int = 3):
    """Return QApplication instance.

    Creates it if it doesn't already exist.

    Parameters
    ----------
    test_time: int
        Time to maintain open the application when testing. It's given in seconds
    """
    import faulthandler

    disable_warnings()
    logger.enable("qtextra")

    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
    if hasattr(Qt, "AA_EnableHighDpiScaling"):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, "AA_UseHighDpiPixmaps"):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    if hasattr(Qt, "AA_UseStyleSheetPropagationInWidgetStyles"):
        QApplication.setAttribute(Qt.AA_UseStyleSheetPropagationInWidgetStyles, True)
    if hasattr(Qt, "AA_ShareOpenGLContexts"):
        QApplication.setAttribute(Qt.AA_ShareOpenGLContexts, True)

    faulthandler.enable()

    Application = MacApplication if sys.platform == "darwin" else QApplication
    app = Application.instance()
    if app is None:
        # Set Application name for Gnome 3
        # https://groups.google.com/forum/#!topic/pyside/24qxvwfrRDs
        app = Application(["qtextra"])

        # Set application name for KDE. See spyder-ide/spyder#2207.
        app.setApplicationName("qtextra")

    test_ci = os.environ.get("TEST_CI_WIDGETS", None)
    if test_ci is not None:
        timer_shutdown = QTimer(app)
        timer_shutdown.timeout.connect(app.quit)
        timer_shutdown.start(test_time * 1000)
    return app


def qframe(
    horz: bool = True,
    with_layout: bool = True,
    add_reload: bool = True,
    set_style: bool = True,
    modules: tuple[str, ...] = DEFAULT_MODULES,
    dev: bool = True,
    toggle: bool = True,
) -> tuple[QApplication, QWidget, QLayout]:
    """Create frame widget."""
    from qtpy import QtWidgets

    app = qapplication()
    frame = QtWidgets.QWidget()
    layout = None
    if with_layout:
        if horz:
            layout = QtWidgets.QHBoxLayout()
            layout.addWidget(theme_toggle_btn(frame))
        else:
            layout = QtWidgets.QVBoxLayout()  # type: ignore[assignment]
        if add_reload:
            w = qdev(modules=modules)
            # w.setMaximumHeight(300)
            layout.addWidget(w)
        frame.setLayout(layout)
    if set_style:
        apply_style(frame)
    if dev:
        from koyo.hooks import install_debugger_hook

        install_debugger_hook()
    if toggle and layout is not None:
        layout.addWidget(theme_toggle_btn(frame))
    return app, frame, layout


def _apply_style_on_widget(widget: QWidget) -> None:
    """Apply stylesheet(s) on the widget."""
    from qtextra.config.theme import THEMES

    THEMES.set_theme_stylesheet(widget)
    print("Applying style on widget")


def apply_style(widget: QWidget) -> None:
    """Apply stylesheet(s) on the widget."""
    from qtextra.config.theme import THEMES

    THEMES.set_theme_stylesheet(widget)
    THEMES.evt_qss_changed.connect(lambda: _apply_style_on_widget(widget))


def qmain(horz: bool = True, set_style: bool = True):
    """Create main widget."""
    from qtpy import QtWidgets

    app = qapplication()
    main = QtWidgets.QMainWindow()
    if horz:
        ha = QtWidgets.QHBoxLayout()
    else:
        ha = QtWidgets.QVBoxLayout()
    main.setCentralWidget(QtWidgets.QWidget())
    main.centralWidget().setLayout(ha)
    if set_style:
        apply_style(main)
    return app, main, ha


def theme_toggle_btn(parent: QWidget):
    """Toggle button."""
    from random import choice

    from qtextra.config import THEMES
    from qtextra.helpers import make_btn

    def _toggle_theme():
        while True:
            theme = choice(THEMES.available_themes())
            if theme != THEMES.theme:
                THEMES.theme = theme
                break
        THEMES.set_theme_stylesheet(parent)

    button = make_btn(parent, "Click here to toggle theme")
    button.clicked.connect(_toggle_theme)
    return button


def get_parent(parent):
    """Get top level parent."""
    if parent is None:
        app = QApplication.instance()
        if app:
            for i in app.topLevelWidgets():
                if isinstance(i, QMainWindow):  # pragma: no cover
                    parent = i
                    break
    return parent


class MacApplication(QApplication):
    """Subclass to be able to open external files with our Mac app."""

    sig_open_external_file = Signal(str)

    def __init__(self, *args):
        QApplication.__init__(self, *args)
        self._never_shown = True
        self._has_started = False
        self._pending_file_open = []
        self._original_handlers = {}

    def event(self, event):
        """Override event handler to catch file open events."""
        if event.type() == QEvent.Type.FileOpen:
            fname = str(event.file())
            if sys.argv and sys.argv[0] == fname:
                # Ignore requests to open own script
                # Later, mainwindow.initialize() will set sys.argv[0] to ''
                pass
            elif self._has_started:
                self.sig_open_external_file.emit(fname)
        return QApplication.event(self, event)



"""Event queue handler."""

from __future__ import annotations

import atexit
import typing as ty
from contextlib import suppress

from loguru import logger
from qtpy.QtCore import QObject, QProcess, Signal  # type: ignore[attr-defined]

from qtextra.queue.cli_qprocess import QProcessWrapper
from qtextra.queue.task import Task
from qtextra.typing import Callback, TaskState
from qtextra.utils.utilities import running_under_pytest

logger = logger.bind(src="Queue")


class CLIQueueHandler(QObject):
    """Simple Queue handler that executes code as its being added to the queue."""

    CAN_FORCE_START = True

    # internally used signal to cancel thread
    _evt_cancel = Signal(Task)
    _evt_pause = Signal(Task, bool)
    # signal to indicate when task had started
    evt_queued = Signal(Task)
    # signal to indicate when task had started
    evt_started = Signal(Task)
    # signal to indicate when task had started
    evt_next = Signal(Task)
    # signal to indicate when task had finished
    evt_finished = Signal(Task)
    # signal to indicate when queue is empty
    evt_empty = Signal()
    # signal to indicate when task had been cancelled
    evt_cancelled = Signal(Task)
    # signal to indicate when task had crashed
    evt_errored = Signal(Task, object)
    evt_part_errored = Signal(Task, object)
    # signal to indicate when task had been paused
    evt_paused = Signal(Task, bool)
    # signal each time there has been a progress update
    evt_progress = Signal(Task)
    # signal to indicate that the queue is closed
    evt_queue_closed = Signal()
    # signal to indicate that all process had been finished
    evt_finished_all = Signal()
    # signal to indicate that a task had been removed
    evt_remove_task = Signal(str)

    def __init__(self, parent: ty.Optional[QObject] = None, n_parallel: int = 1, auto_run: bool = True):
        super().__init__(parent=parent)
        self.auto_run = auto_run
        self.n_parallel = n_parallel

        # flag to close queueing
        self.queue_closed: bool = False

        # dictionary holding currently running thread tasks
        self.active_tasks: ty.Dict[str, QProcessWrapper] = {}
        # pending tasks are added to this queue
        self.pending_queue: ty.List[str] = []
        # running tasks are added to this queue
        self.running_queue: ty.List[str] = []
        # finished tasks are added to this queue
        self.finished_queue: ty.List[str] = []
        self._evt_cancel.connect(self._kill)  # type: ignore[unused-ignore]

        atexit.register(self.close)

    def set_max_parallel(self, n_parallel: int) -> None:
        """Set maximum number of parallel tasks."""
        self.n_parallel = n_parallel
        self.run_queued()

    @property
    def n_tasks(self) -> ty.Tuple[int, int]:
        """Return the number of tasks."""
        return len(self.running_queue), len(self.pending_queue)

    def task_started(self, task: Task) -> None:
        """Triggered whenever function started processing."""
        task_id = task.task_id
        if task_id in self.pending_queue:
            self.pending_queue.remove(task_id)

        if task_id not in self.running_queue:
            self.running_queue.append(task_id)
        worker_obj = self.active_tasks.get(task_id)
        if worker_obj:
            logger.trace(f"Task '{worker_obj.summary()}' started.")
        self.evt_started.emit(task)  # type: ignore[unused-ignore]
        self.run_queued()

    def task_finished(self, task: Task) -> None:
        """Triggered whenever function finished processing."""
        task_id = task.task_id
        worker_obj = self.active_tasks.pop(task_id, None)  # remove from threads
        if worker_obj:
            logger.trace(f"Removed '{task_id}' from active tasks (finished).")
            with suppress(ValueError):
                self.running_queue.remove(task_id)  # remove from restricted queue
            logger.trace(f"Task '{worker_obj.summary()}' ended.")
        else:
            logger.trace(f"Failed to remove '{task_id}' from active tasks (finished).")
        self.finished_queue.append(task_id)
        self.evt_finished.emit(task)  # type: ignore[unused-ignore]
        self.run_queued()

    def task_cancelled(self, task: Task) -> None:
        """Triggered whenever function finished processing."""
        task_id = task.task_id
        worker_obj = self.active_tasks.pop(task_id, None)  # remove from threads
        if worker_obj:
            logger.trace(f"Removed '{task_id}' from active tasks (cancelled).")
            # this condition is always going to be false because we don't allow cancellation of running tasks
            if worker_obj.is_running():
                self.running_queue.remove(task_id)
            else:
                with suppress(ValueError):
                    self.pending_queue.remove(task_id)
            logger.trace(f"Task '{worker_obj.summary()}' was cancelled.")
        else:
            logger.trace(f"Failed to remove '{task_id}' from active tasks (cancelled).")
        self.finished_queue.append(task_id)
        self.evt_cancelled.emit(task)  # type: ignore[unused-ignore]
        self.run_queued()

    def task_errored(self, task: Task) -> None:
        """Triggered whenever function finished processing."""
        task_id = task.task_id
        worker_obj = self.active_tasks.pop(task_id, None)  # remove from threads
        with suppress(ValueError):
            self.running_queue.remove(task_id)
            logger.trace(f"Removed '{task_id}' from running tasks (errored).")
        with suppress(ValueError):
            self.pending_queue.remove(task_id)
            logger.trace(f"Removed '{task_id}' from pending tasks (errored).")
        if worker_obj:
            logger.trace(f"Removed '{task_id}' from active tasks (errored).")
            logger.trace(f"Task '{worker_obj.summary()}' encountered an error")
            error_info = worker_obj.process.readAllStandardOutput().data().decode()
            self.evt_errored.emit(task, error_info)  # type: ignore[unused-ignore]
        else:
            logger.warning(f"Failed to remove task '{task_id}' from active tasks (errored).")
        self.finished_queue.append(task_id)
        self.run_queued()

    def task_part_errored(self, task: Task) -> None:
        """Triggered whenever function finished processing."""
        task_id = task.task_id
        worker_obj = self.active_tasks.get(task_id, None)  # remove from threads
        if worker_obj:
            logger.trace(f"Task '{worker_obj.summary()}' encountered an error")
            error_info = worker_obj.process.readAllStandardOutput().data().decode()
            self.evt_part_errored.emit(task, error_info)  # type: ignore[unused-ignore]

    def _on_finished(self, task_id: str, error_code: int) -> None:
        """Process finished."""
        if error_code == 0:
            with suppress(KeyError):
                worker_obj = self.active_tasks[task_id]
                self.task_finished(worker_obj.task)
        else:
            with suppress(KeyError):
                worker_obj = self.active_tasks[task_id]
                self.task_errored(worker_obj.task)

    def _on_error_change(self, task_id: str, state: QProcess.ProcessError) -> None:
        """Process change of process state."""
        if state == QProcess.ProcessError.FailedToStart:
            logger.error("Task failed to start")
        self._on_finished(task_id, -1)

    def _kill(self, task: Task) -> None:
        """Kill process."""
        try:
            worker_obj = self.active_tasks[task.task_id]
            worker_obj.cancel()
        except KeyError:
            logger.warning(f"Failed to cancel '{task.task_id}'")

    def add_complete_task(self, task: Task) -> None:
        """Emit the 'evt_task_queued' event WITHOUT adding task to the queue."""
        if task.state != TaskState.FINISHED:
            raise ValueError(f"Task '{task.summary()}' is not finished.")
        self.evt_queued.emit(task)  # type: ignore[unused-ignore]

    def add_task(
        self,
        task: Task,
        func_error: ty.Optional[Callback] = None,
        func_start: ty.Optional[Callback] = None,
        func_end: ty.Optional[Callback] = None,
        func_post: ty.Optional[Callback] = None,
        add_delayed: bool = True,
        emit_queued: bool = True,
    ) -> ty.Optional[str]:
        """Adds a task to the queue handler.

        The `Call` handler works by executing consecutive actions.
        1. First, it executes the `func_pre` with No parameters,
        2. Second, it executes the `func` with args and kwargs
            - if action was successful, it will run the `func_result` function with the returned values of the `func`
            - if action was unsuccessful, it will run the `func_error` with error information
        3. Third, it executes the `func_post` with `func_post_args` and `func_post_kwargs` arguments

        The `func_result`, `func_error` and `func_post` are called using the `wx.CallAfter` mechanism to ensure thread
        safety.
        """
        if self.queue_closed:
            self.evt_queue_closed.emit()  # type: ignore[unused-ignore]
            return None
        if task.task_id in self.active_tasks:
            logger.warning(f"Task '{task.summary()}' is already in the queue.")
            return None

        worker_obj = self.make_process(
            task=task,
            func_error=func_error,
            func_start=func_start,
            func_end=func_end,
            func_post=func_post,
        )
        self.add_worker(worker_obj, add_delayed=add_delayed, emit_queued=emit_queued)
        return task.task_id

    def add_worker(self, worker_obj: QProcessWrapper, add_delayed: bool = True, emit_queued: bool = True) -> None:
        """Add call object to the queue.

        Parameters
        ----------
        worker_obj : QProcessWrapper
            QRunnable object
        add_delayed : bool
            Flag to indicate whether the task should be run immediately (if possible) or if it should be added to the
            queue and the request for run is made later.
        emit_queued : bool
            Flag to indicate whether the `evt_queued` event should be emitted.
        """
        if not isinstance(worker_obj, QProcessWrapper):
            raise ValueError("You can only add 'QProcessWrapper' objects to the queue")
        logger.trace(f"Added task='{worker_obj.task_id}' to the queue (queue size {self.count()})")

        # cancel task
        task_id = worker_obj.task_id
        self.active_tasks[task_id] = worker_obj
        if task_id not in self.pending_queue:
            self.pending_queue.append(task_id)
        if emit_queued:
            self.evt_queued.emit(worker_obj.task)  # type: ignore[unused-ignore]

        # Immediately run task if in pytest environment
        if running_under_pytest():
            worker_obj.run()
            logger.trace(f"Running task '{worker_obj.task_id}' in pytest environment")
        else:
            # don't start next task
            if add_delayed or not self.is_available():
                logger.trace("Added task to queue without running it.")
                return
            worker_obj.run()
            logger.trace(f"Running task '{worker_obj.task_id}' immediately.")

    def make_process(
        self,
        task: Task,
        # task_kind: TaskKind = TaskKind.NONE,
        func_error: ty.Optional[Callback] = None,
        func_start: ty.Optional[Callback] = None,
        func_end: ty.Optional[Callback] = None,
        func_post: ty.Optional[Callback] = None,
    ) -> QProcessWrapper:
        """Make QProcess."""
        worker_obj = QProcessWrapper(self, task, func_start, func_error, func_end, func_post)
        worker_obj.evt_started.connect(self.task_started)
        worker_obj.evt_finished.connect(self.task_finished)
        worker_obj.evt_next.connect(self.evt_next.emit)  # type: ignore[unused-ignore]
        worker_obj.evt_errored.connect(self.task_errored)
        worker_obj.evt_part_errored.connect(self.task_part_errored)
        worker_obj.evt_progress.connect(self.evt_progress.emit)  # type: ignore[unused-ignore]
        worker_obj.evt_paused.connect(self.evt_paused.emit)  # type: ignore[unused-ignore]
        worker_obj.evt_cancelled.connect(self.task_cancelled)  # type: ignore[unused-ignore]
        return worker_obj

    def run_queued(self) -> None:
        """Run another object."""
        if self.queue_closed and len(self.running_queue) == 0:
            self.evt_finished_all.emit()  # type: ignore[unused-ignore]
            return

        # start another task
        if self.auto_run and self.is_available() and self.pending_queue:
            task_id = self.pending_queue.pop(0)  # get the first item
            worker_obj = self.active_tasks[task_id]  # get another worker that was queued previously
            worker_obj.run()

        if not self.pending_queue and not self.running_queue:
            self.evt_empty.emit()

    def run_force(self, task: Task) -> None:
        """Run specific task."""
        try:
            logger.trace(f"Manually starting '{task.summary()}'")
            task_id = task.task_id
            worker_obj = self.active_tasks[task_id]  # get another worker that was queued previously
            try:
                worker_obj.run()
            except Exception as e:
                logger.exception(f"Could not forcefully start specified task - '{task.summary()}'")
                self.evt_errored.emit(task, e)
        except (IndexError, KeyError, AttributeError):
            logger.exception(f"Could not forcefully start specified task - '{task.summary()}'")

    @staticmethod
    def can_cancel(_task: Task) -> ty.Tuple[bool, bool]:
        """Get information about thread."""
        return True, True

    @staticmethod
    def can_pause(_task: Task) -> bool:
        """Get information about thread."""
        return False

    def requeue(self, task: Task, remove: bool = False) -> None:
        """Add task to the queue again."""
        if remove:
            self.remove(task)
        if task.task_id in self.active_tasks:
            logger.trace(f"Task '{task.summary()}' is already in the active tasks.")
            self.evt_queued.emit(task)  # type: ignore[unused-ignore]
            return
        if task.task_id in self.pending_queue:
            logger.trace(f"Task '{task.summary()}' is already in the pending queue.")
            return
        self.add_task(task)

    def check_if_in_active(self, task: Task) -> bool:
        """Check whether task is in the active queue."""
        return task.task_id in self.active_tasks

    def check(self, task: Task) -> None:
        """Check whether task is in the pending tasks."""
        if task.task_id in self.active_tasks:
            logger.trace(f"Task '{task.summary()}' is in the active tasks.")
            return
        self.add_task(task, emit_queued=False)
        logger.trace(f"Task '{task.summary()}' is not in the queue.")

    def remove(self, task: str | Task) -> None:
        """Remove a task from the queue."""
        task_id = task if isinstance(task, str) else task.task_id
        if task_id in self.active_tasks:
            self.active_tasks.pop(task_id)
            logger.trace(f"Removed '{task_id}' from active tasks (removed).")
        if task_id in self.pending_queue:
            self.pending_queue.remove(task_id)
            logger.trace(f"Removed '{task_id}' from pending tasks (removed).")

    def cancel(self, task: Task) -> None:
        """Cancel scheduled task."""
        self._evt_cancel.emit(task)  # type: ignore[unused-ignore]

    def pause(self, task: Task, state: bool) -> None:
        """Pause task."""
        try:
            worker_obj = self.active_tasks[task.task_id]
            worker_obj.pause(state)
        except KeyError:
            logger.warning(f"Failed to pause '{task.task_id}'")
        self._evt_pause.emit(task, state)  # type: ignore[unused-ignore]

    def clear(self) -> None:
        """Safely empty queue from waiting tasks."""
        to_remove = [*self.pending_queue]
        for task_id in self.active_tasks:
            if task_id not in to_remove and task_id not in self.running_queue:
                to_remove.append(task_id)

        # cancel all scheduled tasks
        for task_id in to_remove:
            worker_obj = self.active_tasks.pop(task_id, None)
            if worker_obj:
                task = worker_obj.task
                logger.debug(f"Cancelled task '{task.summary()}'")
                with suppress(RuntimeError):
                    self._evt_cancel.emit(task)  # type: ignore[unused-ignore]
        if self.pending_queue:
            self.pending_queue.clear()
            logger.trace("Queue > Cleared queue")
        self.finished_queue.clear()

    def count(self) -> int:
        """Retrieves the count of items in the queue."""
        return len(self.pending_queue)

    def close(self) -> None:
        """Close queue."""
        self.queue_closed = True
        # cancel all scheduled tasks
        self.clear()

    def is_running(self) -> bool:
        """Indicates whether the user can close the window."""
        return len(self.running_queue) > 0

    def is_available(self) -> bool:
        """Indicates whether the queue is busy."""
        return len(self.running_queue) < self.n_parallel

    def is_queued(self, task_id: str) -> bool:
        """Check if task is queued."""
        return task_id in self.pending_queue or task_id in self.running_queue

    def is_finished(self, task_id: str) -> bool:
        """Check if task is finished."""
        return task_id in self.finished_queue

    def remove_task(self, task_id: str) -> None:
        """Remove task from queue."""

        def _remove_if_many(queue: list[str]) -> None:
            while task_id in queue:
                queue.remove(task_id)

        _remove_if_many(self.pending_queue)
        _remove_if_many(self.running_queue)
        _remove_if_many(self.finished_queue)
        with suppress(KeyError):
            worker_obj = self.active_tasks.pop(task_id)
            worker_obj.cancel()
            logger.trace(f"Removed '{task_id}' from active tasks (removed).")
        self.evt_remove_task.emit(task_id)

    def __repr__(self) -> str:
        """Representation of the queue."""
        n_running, n_queued = self.n_tasks
        return f"Queue<queued={n_queued}; running={n_running}>"



"""Color button."""

import typing as ty

import numpy as np
from koyo.color import ColorType, rgbs_to_hex
from koyo.color import transform_color as _transform_color
from qtpy.QtCore import QEvent, QSize, Qt, Signal, Slot
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QColorDialog, QFrame, QPushButton, QWidget

BASE_COLOR = "#FFFFFF"
AnyColorType = ty.Union[ColorType, QColor]
TRANSPARENT = np.array([0, 0, 0, 0], np.float32)


FILL_QSS = """
    QPushButton#color_btn {
        border: none;
        margin: 0px;
        padding: 0px;
        background-color: COLOR;
    }
"""
EMPTY_QSS = FILL_QSS.replace("COLOR", "transparent")
HOVER_QSS = """
    QPushButton#color_btn:hover {
        border: 2px solid black;
    }
"""


class QtColorButton(QPushButton):
    """
    Custom widget to allow selection of color.

    Left-clicking the button shows the color-chooser, while
    right-clicking resets the color to None (no-color).

    Modified from: https://www.mfitzp.com/article/QtColorButton-a-color-selector-tool-for-pyqt/
    """

    # new color
    evt_color_changed: Signal = Signal(str)

    def __init__(self, *args, color=None, size: ty.Tuple[int, int] = (16, 16), **kwargs):
        """
        Parameters
        ----------
        color : str
            valid color
        size : tuple
            size of the button
        *args
            Args to pass to QPushButton
        **kwargs
            Extra kwargs to pass to QPushButton
        """
        self._color = color

        super().__init__(*args, **kwargs)

        self.setFixedSize(QSize(*size))
        self.setObjectName("color_btn")
        self.pressed.connect(self.onColorPicker)

        if self._color is not None:
            self.setColor(self._color)

    @property
    def color(self):
        """Returns the color."""
        return self._color

    def setColor(self, color):
        """Set color."""
        if not hasattr(self, "_color"):
            return

        if color != self._color:
            self._color = color
            self.evt_color_changed.emit(color)

        if self._color:
            self.setStyleSheet(FILL_QSS.replace("COLOR", self._color) + HOVER_QSS)
        else:
            self.setStyleSheet(EMPTY_QSS + HOVER_QSS)

    def onColorPicker(self):
        """Show color-picker dialog to select color.

        Using native dialog by default
        """
        color = QColor(self._color) if self._color is not None else None
        dlg = QColorDialog(color, parent=self)
        if dlg.exec_():
            self.setColor(dlg.currentColor().name())

    def mousePressEvent(self, e):
        """On mouse press."""
        if e.button() == Qt.MouseButton.RightButton:
            self.setColor(None)

        return super().mousePressEvent(e)


class QtColorSwatch(QFrame):
    """A QFrame that displays a color and can be clicked to show a QColorPopup.

    Parameters
    ----------
    parent : QWidget, optional
        parent widget, by default None
    tooltip : Optional[str], optional
        Tooltip when hovering on swatch,
        by default 'click to set color'
    initial_color : ColorType, optional
        initial color, by default will be transparent
    add_frame : bool
        add black frame around the swatch to make it stand out

    Signals
    -------
    evt_color_changed : str
        Emits the new color when the current color changes.
    """

    evt_color_changed = Signal(np.ndarray)

    def __init__(
        self,
        parent: ty.Optional[QWidget] = None,
        tooltip: ty.Optional[str] = None,
        initial_color: ty.Optional[ColorType] = None,
        add_frame: bool = True,
    ):
        super().__init__(parent)
        self.setObjectName("colorSwatch")
        self.setToolTip(tooltip or "Click to set color")
        self.setCursor(Qt.CursorShape.PointingHandCursor)

        self._initial_color = initial_color
        self._color: np.ndarray = TRANSPARENT
        self.evt_color_changed.connect(self._update_swatch_style)
        if initial_color is not None:
            self.set_color(initial_color)
        if add_frame:
            self.setFrameShape(QFrame.Shape.Box)
            self.setLineWidth(2)

    def setEnabled(self, enabled: bool) -> None:
        """Set the enabled state of the widget."""
        super().setEnabled(enabled)
        self.setCursor(Qt.CursorShape.PointingHandCursor if enabled else Qt.CursorShape.ArrowCursor)

    @property
    def color(self) -> np.ndarray:
        """Return the current color."""
        return self._color

    @property
    def hex_color(self) -> str:
        """Return color in HEX code."""
        return rgbs_to_hex(self._color)[0]

    @Slot(np.ndarray)
    def _update_swatch_style(self, _color: ColorType) -> None:
        """Update appearance."""
        rgba = f"rgba({','.join(str(int(x * 255)) for x in self._color)})"
        self.setStyleSheet("#colorSwatch {background-color: " + rgba + ";}")

    def mouseReleaseEvent(self, event: QEvent):
        """Show QColorPopup picker when the user clicks on the swatch."""
        if event.button() == Qt.MouseButton.LeftButton:
            from qtextra.helpers import get_color

            initial = QColor(*(255 * self._color).astype("int"))
            color = get_color(self, initial, False)
            if color is not None:
                self.set_color(color)

    def set_color(self, color: AnyColorType, force: bool = False) -> None:
        """Set the color of the swatch.

        Parameters
        ----------
        color : AnyColorType
            Can be any ColorType recognized by our
            utils.colormaps.standardize_color.transform_color function.
        force : bool
            If True, will emit the color_changed event even if the color
        """
        if isinstance(color, QColor):
            _color = (np.array(color.getRgb()) / 255).astype(np.float32)
        else:
            try:
                from napari.utils.colormaps.standardize_color import transform_color

                _color = transform_color(color)[0]
            except ValueError:
                return self.evt_color_changed.emit(self._color)
            except ImportError:
                try:
                    _color = _transform_color(color)[0]
                except ValueError:
                    return self.evt_color_changed.emit(self._color)

        emit = np.any(self._color != _color)
        self._color = _color
        if emit or np.all(_color == TRANSPARENT):
            self.evt_color_changed.emit(_color)
        if force:
            self._update_swatch_style(_color)

    # Alias methods to offer Qt-like interface
    hexColor = hex_color
    setColor = set_color


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setLayout(ha)
    frame.setMinimumSize(400, 400)

    w = QtColorButton(size=(64, 64), color="#FF00FF")
    ha.addWidget(w)
    ww = QtColorSwatch(initial_color="#FF0000")
    ha.addWidget(ww)
    ww = QtColorSwatch(initial_color="#FFFF00")
    ha.addWidget(ww)
    ww = QtColorSwatch(initial_color=(255, 123, 32))
    ha.addWidget(ww)

    frame.show()
    sys.exit(app.exec_())



"""tqdm-based progress manager."""

import typing as ty

from tqdm import tqdm

try:
    from napari.utils.events.event import EmitterGroup, Event
except ImportError:
    raise ImportError("please install napari using 'pip install napari'") from None


class Progress(tqdm):
    """This class inherits from tqdm and provides an interface for
    progress bars in the napari viewer. Progress bars can be created
    directly by wrapping an iterable or by providing a total number
    of expected updates.

    While this interface is primarily designed to be displayed in
    the viewer, it can also be used without a viewer open, in which
    case it behaves identically to tqdm and produces the progress
    bar in the terminal.

    See tqdm.tqdm API for valid args and kwargs:
    https://tqdm.github.io/docs/tqdm/

    Examples
    --------
    >>> def long_running(steps=10, delay=0.1):
    ...     for i in Progress(range(steps)):
    ...         sleep(delay)

    it can also be used as a context manager:

    >>> def long_running(steps=10, repeats=4, delay=0.1):
    ...     with progress(range(steps)) as pbr:
    ...         for i in pbr:
    ...             sleep(delay)

    or equivalently, using the `progrange` shorthand

    .. code-block:: python

        with progrange(steps) as pbr:
            for i in pbr:
                sleep(delay)

    For manual updates:

    >>> def manual_updates(total):
    ...     pbr = Progress(total=total)
    ...     sleep(10)
    ...     pbr.setDescription("Step 1 Complete")
    ...     pbr.update(1)
    ...     # must call pbr.close() when using outside for loop
    ...     # or context manager
    ...     pbr.close()

    """

    monitor_interval = 0  # set to 0 to disable the thread

    def __init__(
        self,
        iterable: ty.Optional[ty.Iterable] = None,
        desc: ty.Optional[str] = None,
        total: ty.Optional[int] = None,
        bar_format: str = "|{bar}| {n_fmt}/{total_fmt} [ETA: {remaining}/{elapsed} {rate_fmt}]",
        *args: ty.Any,
        **kwargs: ty.Any,
    ) -> None:
        self.events = EmitterGroup(value=Event, description=Event, overflow=Event, eta=Event, close=Event)
        self.is_init = True
        super().__init__(iterable, *args, desc=desc, total=total, bar_format=bar_format, **kwargs)

        if not self.desc:
            self.set_description("")
        self.is_init = False

    def __repr__(self) -> str:
        """Return description."""
        return self.desc

    def display(self, msg: ty.Optional[str] = None, pos: ty.Optional[int] = None) -> None:
        """Update the display and emit eta event."""
        # just plain tqdm if we don't have gui
        if not self.gui and not self.is_init:
            super().display(msg, pos)
            return
        etas = ""
        if self.total != 0:  # type: ignore[has-type]
            etas = str(self).split("|")[-1]
        self.events.eta(value=etas)

    def update(self, n: ty.Optional[float] = None) -> None:
        """Update progress value by n and emit value event."""
        super().update(n)
        self.events.value(value=self.n)

    def increment_with_overflow(self) -> None:
        """Update if not exceeding total, else set indeterminate range."""
        if self.n == self.total:  # type: ignore[has-type]
            self.total = 0
            self.events.overflow()
        else:
            self.update(1)

    def set_description(self, desc: str) -> None:
        """Update progress description and emit description event."""
        super().set_description(desc, refresh=True)
        self.events.description(value=desc)

    def close(self) -> None:
        """Close progress object and emit event."""
        if self.disable:
            return
        self.events.close()
        self.events.disconnect()
        super().close()



"""Progress bar with label."""

from __future__ import annotations

from napari.utils.events import Event
from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QHBoxLayout, QLabel, QProgressBar, QVBoxLayout, QWidget

from qtextra.utils.progress import Progress


class QtLabeledProgressBar(QWidget):
    """QProgressBar with QLabels for description and ETA."""

    def __init__(self, parent: QWidget | None = None, progress: Progress | None = None) -> None:
        super().__init__(parent)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)

        self.progress = progress

        self.description_label = QLabel()
        self.qt_progress_bar = QProgressBar()
        self.qt_progress_bar.setTextVisible(True)

        layout = QHBoxLayout()
        layout.addWidget(self.description_label)
        layout.addWidget(self.qt_progress_bar)

        pbar_layout = QVBoxLayout(self)
        pbar_layout.addLayout(layout, stretch=True)
        pbar_layout.setContentsMargins(0, 0, 0, 0)
        pbar_layout.setSpacing(0)

        self.setMinimum = self.qt_progress_bar.setMinimum
        self.setMaximum = self.qt_progress_bar.setMaximum

    def setRange(self, min_value: int, max_value: int) -> None:
        """Set range."""
        self.qt_progress_bar.setRange(min_value, max_value)

    def setValue(self, value: int) -> None:
        """Set value."""
        self.qt_progress_bar.setValue(value)
        QApplication.processEvents()

    def setDescription(self, value: str) -> None:
        """Set description."""
        self.description_label.setText(value)
        QApplication.processEvents()

    def _set_value(self, event: Event) -> None:
        self.setValue(event.value)

    def _get_value(self) -> int:
        return self.qt_progress_bar.value()

    def _set_description(self, event: Event) -> None:
        self.setDescription(event.value)

    def _make_indeterminate(self, event: Event) -> None:
        self.setRange(0, 0)

    def _set_eta(self, event: Event) -> None:
        self.qt_progress_bar.setFormat(event.value)

    def _on_clear(self, event: Event) -> None:
        self.description_label.setText("")


def set_progress_bar(progress: Progress, progress_bar: QtLabeledProgressBar):
    """Make progress bar."""
    progress.gui = True
    progress.leave = False

    # connect progress object events to updating progress bar
    progress.events.value.connect(progress_bar._set_value)
    progress.events.description.connect(progress_bar._set_description)
    progress.events.overflow.connect(progress_bar._make_indeterminate)
    progress.events.eta.connect(progress_bar._set_eta)
    progress.events.close.connect(progress_bar._on_clear)

    # set its range etc. based on progress object
    if progress.total is not None:
        progress_bar.setRange(progress.n, progress.total)
        progress_bar.setValue(progress.n)
    else:
        progress_bar.setRange(0, 0)
        progress.total = 0
    progress_bar.setDescription(progress.desc)


def _test(pbar):
    import time

    prog = Progress(range(50))
    set_progress_bar(prog, pbar)
    for _v in prog:
        time.sleep(0.1)
    prog.close()


if __name__ == "__main__":  # pragma: no cover
    import sys
    from functools import partial

    from qtpy.QtWidgets import QPushButton

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setLayout(ha)
    frame.setMinimumSize(400, 400)

    pbar1 = QtLabeledProgressBar()
    ha.addWidget(pbar1)

    pbar2 = QtLabeledProgressBar()
    ha.addWidget(pbar2)

    btn = QPushButton("Press me to start")
    btn.clicked.connect(partial(_test, pbar1))
    btn.clicked.connect(partial(_test, pbar2))
    ha.addWidget(btn)

    frame.show()
    sys.exit(app.exec_())



"""Changelog dialog."""

from __future__ import annotations

import typing as ty
from contextlib import suppress
from pathlib import Path
from urllib import request

from koyo.compression import unzip_directory
from koyo.typing import PathLike
from qtpy.QtWidgets import QFormLayout, QLabel, QProgressBar, QPushButton, QTextEdit, QWidget
from superqt.utils import create_worker, ensure_main_thread, qthrottled
from tqdm import tqdm

import qtextra.helpers as hp
from qtextra.config import THEMES
from qtextra.widgets.qt_code_widget import Codelighter
from qtextra.widgets.qt_dialog import QtFramelessTool


class ChangelogDialog(QtFramelessTool):
    """Changelog."""

    HIDE_WHEN_CLOSE = False

    def __init__(
        self,
        parent: ty.Optional[QWidget],
        text: str,
        language: str = "markdown",
        download_url: ty.Optional[str] = None,
        download_info: ty.Optional[str] = None,
        path_to_file: ty.Optional[PathLike] = None,
    ) -> None:
        self.text = text
        self.language = language
        self.download_url = download_url
        self.download_info = download_info or ""
        self.path_to_file = path_to_file
        self.path_to_out = get_path(path_to_file) if path_to_file else None

        super().__init__(parent)
        self.setMinimumWidth(600)
        self.setMinimumHeight(400)

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QFormLayout:
        """Make panel."""
        self.text_edit = QTextEdit()
        self._highlight = Codelighter(self.text_edit.document(), THEMES.syntax_style, self.language)
        self.text_edit.setText(self.text)
        self.text_edit.setReadOnly(True)
        self.download_label = hp.make_label(self, self.download_info, enable_url=True)
        self.download_btn = QPushButton("Download")
        self.download_btn.clicked.connect(self.on_download)
        self.progress_bar = QProgressBar()
        self.progress_bar.hide()
        self.progress_label = hp.make_label(self, enable_url=True)
        self.progress_label.hide()
        if self.download_url is None:
            self.download_label.hide()
            self.download_btn.hide()
        if self.path_to_out is not None and self.path_to_out.exists():
            self.progress_label.setText(
                f"<a href='{self.path_to_out.as_uri()}'>Open {self.path_to_out} - it's been previously downloaded.</a>"
            )
            self.progress_label.show()
            self.download_btn.hide()

        layout = hp.make_form_layout(parent=self)
        layout.addRow(self._make_close_handle("Changelog")[1])
        layout.addRow(self.text_edit)
        layout.addRow(self.download_label)
        layout.addRow(self.download_btn)
        layout.addRow(self.progress_bar)
        layout.addRow(self.progress_label)
        layout.addRow(self.progress_label)
        return layout

    # @ensure_main_thread()
    def on_download(self, *_args: ty.Any) -> None:
        """Download."""
        if self.download_url is None:
            return

        pbar = tqdm(unit="B", unit_scale=True, unit_divisor=1024, desc="Downloading...", mininterval=1)
        reporthook = report_hook(self.progress_bar, self.progress_label, pbar)
        create_worker(
            download_file,
            self.download_url,
            # self.progress_bar,
            # self.progress_label,
            self.path_to_file,
            reporthook,
            pbar,
            _start_thread=True,
            _connect={
                "returned": self._on_download_finished,
                "yielded": self._on_download_yielded,
                "started": self._on_download_started,
            },
        )

    @ensure_main_thread()
    def _on_download_yielded(self, text: str) -> None:
        self.progress_label.setText(text)
        self.progress_label.show()

    @ensure_main_thread()
    def _on_download_started(self) -> None:
        """Download start."""
        self.progress_bar.show()
        self.progress_label.show()
        self.download_btn.hide()

    @ensure_main_thread()
    def _on_download_finished(self, res: tuple[tqdm, str]) -> None:
        """Download finished."""
        pbar, text = res
        pbar.close()
        self.progress_bar.hide()
        self.download_btn.hide()
        self.progress_label.setText(text)
        self.progress_label.show()


def get_path(path_to_file: PathLike) -> Path:
    """Get formatted path."""
    path_to_file = Path(path_to_file)
    if path_to_file.suffix == ".zip":
        return path_to_file.parent / path_to_file.stem
    return path_to_file


def report_hook(
    progress_bar: QProgressBar, label: QLabel, pbar: tqdm
) -> ty.Callable[[int, int, ty.Optional[int]], None]:
    """Download progress."""
    last_b = [0]

    @ensure_main_thread()
    @qthrottled(timeout=500)
    def update_to(b: int = 1, bsize: int = 1, tot_size: ty.Optional[int] = None) -> None:
        """Update progress.

        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tot_size  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        with suppress(RuntimeError):
            if tot_size is not None:
                progress_bar.setMaximum(tot_size)
                progress_bar.setValue(0)
                pbar.total = tot_size
            pbar.update((b - last_b[0]) * bsize)
            progress_bar.setValue(pbar.n)
            text = f"{pbar.n / 1024**2:.2f} / {pbar.total / 1024**2:.2f} MB"
            label.setText(text)
        last_b[0] = b

    return update_to  # type: ignore[return-value]


def download_file(
    url: str, path_to_file: PathLike, reporthook: ty.Callable, pbar: tqdm
) -> ty.Generator[str, None, None] | tuple[tqdm, str]:
    """Download data."""
    path_to_file = Path(path_to_file)
    if not path_to_file.exists():
        # create report hook
        temp_file = path_to_file.parent / (path_to_file.name + ".temp")
        _, _ = request.urlretrieve(url, temp_file, reporthook=reporthook)
        temp_file.rename(path_to_file)
    if path_to_file.suffix == ".zip":
        yield "Unzipping..."
        path_to_out = get_path(path_to_file)
        unzip_directory(path_to_file, path_to_out, remove_archive=True)
        yield "Unzipped"
    else:
        path_to_out = path_to_file
    return pbar, f"<a href='{path_to_out.as_uri()}' style='color: blue;'>Open {path_to_out}</a>"



"""Status bar widgets.

Taken and modified from spyder.widgets.status
"""

from __future__ import annotations

import typing as ty

from qtpy.QtCore import QSize, Qt, QTimer, Signal
from qtpy.QtGui import QFont, QIcon, QMouseEvent
from qtpy.QtWidgets import QHBoxLayout, QLabel, QProgressBar, QSizePolicy, QStatusBar, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_button_tool import QtToolButton
from qtextra.widgets.qt_label_icon import QtQtaLabel


class QtStatusbarLabel(QLabel):
    """Statusbar label."""

    def __init__(self, parent: QWidget | None, statusbar: QStatusBar, text: str = "", font_size: int = 7):
        """Status bar progress bar."""
        super().__init__(parent)

        self.setText(text)
        self.setFont(hp.get_font(font_size))

        # Setup
        statusbar.addPermanentWidget(self)


class QtStatusbarToolBtn(QtToolButton):
    """Status bar tool button base."""

    def __init__(self, parent: QWidget | None, statusbar: QStatusBar, text: str = "", icon: QIcon | None = None):
        """Status bar widget base."""
        super().__init__(parent, text=text, icon=icon)
        self.setFont(hp.get_font(7))

        # Widget
        self._status_bar = statusbar
        # Setup
        statusbar.addPermanentWidget(self)


class QtStatusbarProgressbar(QProgressBar):
    """Statusbar progressbar."""

    def __init__(self, parent: QWidget | None, statusbar: QStatusBar):
        """Status bar progress bar."""
        super().__init__(parent)

        # Setup sizing policy
        size_policy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        size_policy.setHorizontalStretch(0)
        size_policy.setVerticalStretch(0)
        size_policy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
        self.setSizePolicy(size_policy)

        self.setFont(hp.get_font(7))

        # Setup
        statusbar.addPermanentWidget(self)


class QtStatusbarSpinner(QWidget):
    """Statusbar spinner."""

    def __init__(self, parent: QWidget | None, statusbar: QStatusBar):
        super().__init__(parent)

        self.spinner, self.movie = hp.make_loading_gif(parent, size=(18, 18))

        layout = QHBoxLayout()
        layout.addWidget(self.spinner)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        statusbar.addPermanentWidget(self)

    def start(self) -> None:
        """Start spinner."""
        self.spinner.show()

    def stop(self) -> None:
        """Stop spinner."""
        self.spinner.hide()


class QtStatusbarIconWidget(QWidget):
    """Status bar widget with single icon."""

    # Signals
    evt_clicked = Signal()

    def __init__(self, parent: QWidget | None, statusbar: QStatusBar, name: str = "", index: ty.Optional[int] = None):
        """Status bar widget base."""
        super().__init__(parent)
        self.setMouseTracking(True)

        # Widget
        self._status_bar = statusbar
        self._icon = None
        self._pixmap = None
        self._icon_size = QSize(16, 16)  # ; should this be adjutable?
        self.label_icon = QtQtaLabel()
        if name:
            self.label_icon.set_qta(name)
            self.label_icon.set_small()

        # Layout setup
        layout = QHBoxLayout(self)
        layout.setSpacing(0)  # Reduce space between icon and label
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.label_icon, alignment=Qt.AlignmentFlag.AlignHCenter)

        # Setup
        if index is None:
            statusbar.addPermanentWidget(self)
        else:
            statusbar.insertPermanentWidget(index, self)
        self.update_tooltip()

    def update_tooltip(self) -> None:
        """Update tooltip for widget."""
        tooltip = self.get_tooltip()
        if tooltip:
            if self.label_icon:
                self.label_icon.setToolTip(tooltip)
            self.setToolTip(tooltip)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:  # type: ignore[override]
        """Override Qt method to allow for click signal."""
        super().mousePressEvent(event)
        self.evt_clicked.emit()

    def get_tooltip(self) -> str:
        """Return the widget tooltip text."""
        return ""

    def flash(self, duration: int = 500, color: tuple = (0.5, 0.5, 0.5, 0.5)) -> None:
        """Add simple flash animation to highlight an event."""
        hp.add_flash_animation(self, duration, color, 3)


class QtStatusbarWidget(QtStatusbarIconWidget):
    """Status bar widget base."""

    # Signals
    evt_clicked = Signal()

    def __init__(
        self,
        parent: QWidget | None,
        statusbar: QStatusBar,
        name: str = "",
        example_text: str | None = None,
        hide_label: bool = False,
    ):
        """Status bar widget base."""
        super().__init__(parent, statusbar, name=name)
        # Variables
        self.value: str | None = None
        self.label_value = QLabel()

        # Layout setup
        self.layout().addWidget(self.label_value)  # type: ignore[union-attr]

        # See spyder-ide/spyder#9044.
        self.label_value.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        self.label_value.setFont(hp.get_font(7, QFont.Weight.Bold))
        if hide_label:
            self.label_value.setVisible(False)

        # Widget setup
        if example_text:
            fm = self.label_value.fontMetrics()
            self.label_value.setMinimumWidth(fm.horizontalAdvance(example_text))
        self.setMinimumWidth(self.label_icon.minimumWidth() + self.label_value.minimumWidth())

        # Setup
        statusbar.addPermanentWidget(self)
        self.set_value("")
        self.update_tooltip()

    def set_value(self, value: str) -> None:
        """Set formatted text value."""
        self.value = value
        self.label_value.setText(value)

    def update_tooltip(self) -> None:
        """Update tooltip for widget."""
        tooltip = self.get_tooltip()
        if tooltip and hasattr(self, "label_value"):
            self.label_value.setToolTip(tooltip)
            if self.label_icon:
                self.label_icon.setToolTip(tooltip)
            self.setToolTip(tooltip)


class QtStatusbarTimerBase(QtStatusbarWidget):
    """Status bar widget base for widgets that update based on timers."""

    def __init__(self, parent: QWidget | None, statusbar: QStatusBar, name: str = "", example_text: str | None = None):
        """Status bar widget base for widgets that update based on timers."""
        self.timer = None  # Needs to come before parent call
        super().__init__(parent, statusbar, name=name, example_text=example_text)
        self._interval = 2000

        # Setup
        if self.is_supported():
            self.timer = QTimer(self)
            self.timer.timeout.connect(self.update_status)
            self.timer.start(self._interval)
        else:
            self.hide()

    def setVisible(self, value: bool) -> None:
        """Override Qt method to stops timers if widget is not visible."""
        if self.timer is not None:
            if value:
                self.timer.start(self._interval)
            else:
                self.timer.stop()
        super().setVisible(value)

    def is_supported(self) -> bool:
        """Return True if feature is supported."""
        try:
            self.import_test()
            return True
        except ImportError:
            return False

    def update_status(self) -> None:
        """Update status label widget, if widget is visible."""
        if self.isVisible():
            self.label_value.setText(self.get_value())

    def set_interval(self, interval: int) -> None:
        """Set timer interval (ms)."""
        self._interval = interval
        if self.timer is not None:
            self.timer.setInterval(interval)

    def import_test(self) -> None:
        """Raise ImportError if feature is not supported."""
        raise NotImplementedError

    def get_value(self) -> str:
        """Return formatted text value."""
        raise NotImplementedError


class QtStatusbarMemory(QtStatusbarTimerBase):
    """Status bar widget for system memory usage."""

    def import_test(self) -> None:
        """Raise ImportError if feature is not supported."""
        from qtextra.utils.utilities import memory_usage

        del memory_usage

    def get_value(self) -> str:
        """Return memory usage."""
        from qtextra.utils.utilities import memory_usage

        text = "%d%%" % memory_usage()
        return "Mem " + text.rjust(3) if not self._pixmap else " " + text.rjust(3)

    def get_tooltip(self) -> str:
        """Return the widget tooltip text."""
        return "Memory usage"


class QtStatusbarProcessMemory(QtStatusbarTimerBase):
    """Status bar widget for system memory usage."""

    def import_test(self) -> None:
        """Raise ImportError if feature is not supported."""
        from qtextra.utils.utilities import process_memory_usage

        del process_memory_usage

    def get_value(self) -> str:
        """Return memory usage."""
        from qtextra.utils.utilities import process_memory_usage

        try:
            text = "%d%%" % process_memory_usage()
        except Exception:
            text = "N/A"
        return "Mem " + text.rjust(3) if not self._pixmap else " " + text.rjust(3)

    def get_tooltip(self) -> str:
        """Return the widget tooltip text."""
        return "Memory usage"


class QtStatusbarCPU(QtStatusbarTimerBase):
    """Status bar widget for system cpu usage."""

    def import_test(self) -> None:
        """Raise ImportError if feature is not supported."""

    def get_value(self) -> str:
        """Return CPU usage."""
        import psutil

        text = "%d%%" % psutil.cpu_percent(interval=0)
        return "CPU " + text.rjust(3) if not self._pixmap else " " + text.rjust(3)

    def get_tooltip(self) -> str:
        """Return the widget tooltip text."""
        return "CPU usage"


class QtStatusbarProcessCPU(QtStatusbarTimerBase):
    """Status bar widget for system cpu usage."""

    def import_test(self) -> None:
        """Raise ImportError if feature is not supported."""

    def get_value(self) -> str:
        """Return CPU usage."""
        import psutil

        try:
            text = "%d%%" % psutil.Process().cpu_percent(interval=0)
        except Exception:
            text = "N/A"
        return "CPU " + text.rjust(3) if not self._pixmap else " " + text.rjust(3)

    def get_tooltip(self) -> str:
        """Return the widget tooltip text."""
        return "CPU usage"



from __future__ import annotations

import typing as ty

from pydantic import BaseModel, ConfigDict, validator
from qtpy.QtCore import Qt
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QFrame, QVBoxLayout, QWidget

import qtextra.helpers as hp


class Tile(BaseModel):
    """Workflow model."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    title: str
    description: str
    icon: ty.Optional[str] = ""
    func: ty.Optional[ty.Callable] = None
    icon_kws: ty.Optional[ty.Dict[str, ty.Any]] = None
    warning: str = ""

    @validator("func", pre=True)
    def _validate_widget(cls, value) -> ty.Callable:
        """Check correct model is provided."""
        if not value:
            return lambda: None
        if not callable(value):
            raise ValueError("Widget class must have a run method.")
        return value


class QtTileWidget(QFrame):
    """Widget to display workflow."""

    def __init__(self, parent: QWidget, tile: Tile):
        super().__init__(parent)
        self.setMouseTracking(True)
        self.setFixedSize(275, 275)
        self.setFrameShape(QFrame.Shape.Box)

        self._tile = tile
        # layout: Title, image, description
        self._title = hp.make_label(
            self, self._tile.title, wrap=True, alignment=Qt.AlignmentFlag.AlignHCenter, object_name="large_text"
        )
        self._title.adjustSize()
        if self._tile.icon:
            self._image = hp.make_qta_label(self, self._tile.icon, **(self._tile.icon_kws or {}))
            self._image.set_xxxlarge()
        else:
            self._image = hp.make_label(self, "")  # type: ignore[assignment]
        self._description = hp.make_label(
            self, self._tile.description, wrap=True, alignment=Qt.AlignmentFlag.AlignHCenter
        )
        hp.set_retain_hidden_size_policy(self._description)

        self._warning = hp.make_label(
            self, self._tile.warning, wrap=True, alignment=Qt.AlignmentFlag.AlignHCenter, object_name="small_text"
        )
        hp.set_retain_hidden_size_policy(self._warning)
        if not self._tile.warning:
            self._warning.hide()

        layout = QVBoxLayout(self)
        layout.setSpacing(2)
        layout.setContentsMargins(2, 2, 2, 2)
        layout.addLayout(
            hp.make_h_layout(self._title, spacing=0, stretch_id=0, alignment=Qt.AlignmentFlag.AlignHCenter),
            stretch=1,
        )
        layout.addWidget(self._image, stretch=2, alignment=Qt.AlignmentFlag.AlignCenter)
        layout.addLayout(
            hp.make_h_layout(self._description, spacing=0, stretch_id=0, alignment=Qt.AlignmentFlag.AlignHCenter),
            stretch=1,
        )
        layout.addLayout(
            hp.make_h_layout(self._warning, spacing=0, stretch_id=0, alignment=Qt.AlignmentFlag.AlignHCenter),
        )
        layout.addStretch(1)

    def mousePressEvent(self, event: QMouseEvent) -> None:  # type: ignore[override]
        """Handle mouse press event."""
        self._tile.func()
        return super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:  # type: ignore[override]
        """Handle mouse release event."""
        return super().mouseReleaseEvent(event)


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe()
        frame.setMinimumSize(400, 400)

        model = Tile(
            title="One-vs-One",
            description="Compare intensity of one ion to another.",
            func=lambda: print("Hello"),
            icon="mdi.chart-scatter-plot",
        )
        widget = QtTileWidget(frame, model)
        ha.addWidget(widget)
        widget = QtTileWidget(frame, model)
        ha.addWidget(widget)
        model = Tile(
            title="One-vs-One",
            description="Convert multi-scene CZI images or other formats to OME-TIFF.",
            func=lambda: print("Hello"),
            icon="mdi.chart-scatter-plot",
            icon_kws={"color": "#ff0000"},
            warning="<i>Not available on Apple Silicon due to a bug I can't find...</i>",
        )
        widget = QtTileWidget(frame, model)
        ha.addWidget(widget)
        frame.show()
        sys.exit(app.exec_())

    _main()  # type: ignore[no-untyped-call]



"""Eliding label."""

from qtpy.QtCore import Qt
from superqt import QElidingLabel


class QtElidingLabel(QElidingLabel):
    """A single-line eliding QLabel."""

    def __init__(
        self,
        text="",
        bold: bool = False,
        parent=None,
        elide: Qt.TextElideMode = Qt.TextElideMode.ElideMiddle,
        multiline: bool = False,
    ):
        super().__init__(text, parent=parent)
        self.setElideMode(elide)
        self.setWordWrap(multiline)
        font = self.font()
        font.setBold(bold)
        self.setFont(font)

    def full_text(self) -> str:
        """Get full text."""
        return self._text


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe(False)

        widget = QtElidingLabel(
            parent=frame,
            text="This is a lot of text that should be cut because its way too long for such a short label\n" * 5,
            multiline=True,
        )

        ha.addWidget(widget)
        frame.show()
        sys.exit(app.exec_())

    _main()



"""Animation base."""

import os.path
from contextlib import contextmanager
from enum import auto

from napari.utils.misc import StringEnum
from qtpy.QtCore import QObject, QTimer, Signal, Slot


class LoopMode(StringEnum):
    """Looping mode for animating an axis.

    LoopMode.ONCE
        Animation will stop once movie reaches the max frame (if fps > 0) or the first frame (if fps < 0).
    LoopMode.LOOP
        Movie will return to the first frame after reaching the last frame, looping continuously until stopped.
    LoopMode.BACK_AND_FORTH
        Movie will loop continuously until stopped, reversing direction when the maximum or minimum frame has been
         reached.
    """

    ONCE = auto()
    LOOP = auto()
    BACK_AND_FORTH = auto()


LOOP_COUNTS = {
    LoopMode.ONCE: 1,
    LoopMode.LOOP: 1,
    LoopMode.BACK_AND_FORTH: 2,
}
LOOP_MODE_TRANSLATIONS = {LoopMode.ONCE: "Once", LoopMode.LOOP: "Loop", LoopMode.BACK_AND_FORTH: "Back-and-forth"}


class QtAnimationBase(QObject):
    """Animation base.

    This class is responsible for animating/progressing through set of frames. This can include different ion images,
    mass spectra, m/z vs dt heatmaps or different ion mobility bins. The class should not be used directly, but rather
    subclassed and the `setup` and `update` methods must be implemented.

    Attributes
    ----------
    timer : QTimer
        Instance of the timer that takes care of updates.
    is_running : bool
        Flag to indicate whether the animator is running.
    interval : int
        Interval between each update (in ms)
    step : int
        Step size - this determines whether an update should occur on subsequent frames or some  frames should be
        skipped.
    current : int
        Current frame.
    min_point : int
        Minimum value to start animation from.
    max_point : int
        Maximum value where the animation should stop or reset.
    loop_mode : LoopMode
        Enum that determines what should happen when animation reaches its final frame. See `LoopMode` for available
        modes.

    Private Attributes
    ------------------
    _stop : bool
        Flag to indicate that animation was or should be stopped.
    """

    # event to lock certain ui elements when an action is taking place
    evt_lock = Signal(bool)
    # triggered when animation had started
    evt_started = Signal()
    # triggered when animation is finished
    evt_finished = Signal()
    # triggered whenever class is reset
    evt_reset = Signal()
    # triggered whenever animation was started or stopped
    evt_active = Signal(bool)
    # triggered when next frame is requested
    evt_next_requested = Signal(int)

    # triggered when writer has been setup
    evt_writer_setup = Signal()
    # triggered when writer had been updated
    evt_writer_updated = Signal(int, int)
    # triggered when writer had been closed
    evt_writer_finished = Signal()

    def __init__(self, parent, fps):
        QObject.__init__(self, parent)
        # timer
        self.timer = QTimer(self)

        self._stop: bool = False  # flag to indicate to stop the animation
        self.is_running: bool = False
        self.interval: int = 0
        self.step: int = 1
        self.current, self.min_point, self.max_point = 0, None, None
        self.loop_mode = str(LoopMode.ONCE)
        self.dim_range = (0, 0, 1)

        # writer attributes
        self._writer = None
        self._writer_n_loops: int = 1
        self._writer_current_iter: int = 0
        self._writer_max_iter: int = 0
        self._writer_finished: bool = False
        self._writer_current_step: int = 0
        self._writer_max_step: int = 0

    def reset(self):
        """Reset."""
        self.is_running: bool = False
        self.interval: int = 0
        self.step: int = 1
        self.current, self.min_point, self.max_point = 0, None, None
        self.loop_mode = str(LoopMode.ONCE)
        self.dim_range = (0, 0, 1)

        # writer attributes
        self._writer = None
        self._writer_n_loops: int = 1
        self._writer_current_iter: int = 0
        self._writer_max_iter: int = 0
        self._writer_finished: bool = False
        self._writer_current_step: int = 0
        self._writer_max_step: int = 0

    def setup(self, *args, **kwargs):
        """Setup."""
        raise NotImplementedError("Must implement method")

    def update(self, *args, **kwargs):
        """Update."""
        raise NotImplementedError("Must implement method")

    def _animate(self, current: int):
        """Yield animation data."""

    @property
    def writer(self):
        """Get handle to the writer."""
        return self._writer

    @property
    def writer_max_iter(self) -> int:
        """Get maximum number of iterations the loop should go on for."""
        return self._writer_max_iter

    @writer_max_iter.setter
    def writer_max_iter(self, value):
        self._writer_max_iter = value
        self._writer_current_iter = value
        self._writer_current_step = 0
        self._writer_max_step = value * self.max_point

    def set_writer(
        self,
        filename: str,
        fps: int = 10,
        loop: int = 0,
        duration: float = 0.1,
        n_loops: int = 1,
        quality: int = 5,
        **kwargs,
    ):
        """Set writer where canvas data can be written to.

        Parameters
        ----------
        filename : str
            Path to the filename where the writer can write to
        fps : int
            Frames per second to be written to the file
        loop : int
            Flag to indicate how many loops should be written to the file. Value of 0 means that gif will
            loop forever.
        duration : float
            Time (in seconds) between each frame.
        n_loops : int
            The number of iterations that should be written to the file
        quality : int
            Number from 1 (lowest quality) to 9. Only applies to non-gif extensions.
        **kwargs
            Additional keyword arguments
        """
        import imageio

        if n_loops < 1:
            raise ValueError("The `n_loops` value must be larger than 1.")
        if fps < 1:
            raise ValueError("The `fps` value must be larger than 1.")

        if self._writer is not None:
            self.close_writer()

        if filename.endswith(".gif"):
            self._writer = imageio.get_writer(filename, mode="I", duration=duration, fps=fps, loop=loop)
        elif os.path.splitext(filename)[-1] in [".mov", ".avi", ".mpg", ".mpeg", ".mp4", ".mkv", ".wmv"]:
            self._writer = imageio.get_writer(filename, mode="I", fps=fps, quality=quality)
        self._writer_n_loops = n_loops
        self.writer_max_iter = LOOP_COUNTS[self.loop_mode] * n_loops
        self._writer_finished = False
        self.set_writer_extras(**kwargs)
        self.evt_writer_setup.emit()

    def set_writer_extras(self, **kwargs):
        """Set any extra parameters that the writer might require."""

    def close_writer(self):
        """Close writer."""
        if self._writer:
            self._writer.close()
            self._writer = None
            self._writer_current_iter = 0

    @Slot(float)
    def set_fps(self, fps: float):
        """Set the frames per second value for the animation.

        Parameters
        ----------
        fps : float
            Frames per second for the animation.
        """
        if fps == 0:
            return self.finish()
        self.step = 1 if fps > 0 else -1  # negative fps plays in reverse
        self.interval = 1000 / abs(fps)

    @Slot(str)
    def set_loop_mode(self, mode: str):
        """Set the loop mode for the animation.

        Parameters
        ----------
        mode : str
            Loop mode for animation.
            Available options for the loop mode string enumeration are:
            - LoopMode.ONCE
                Animation will stop once movie reaches the max frame
                (if fps > 0) or the first frame (if fps < 0).
            - LoopMode.LOOP
                Movie will return to the first frame after reaching
                the last frame, looping continuously until stopped.
            - LoopMode.BACK_AND_FORTH
                Movie will loop continuously until stopped,
                reversing direction when the maximum or minimum frame
                has been reached.
        """
        self.loop_mode = LoopMode(mode)

    def start(self):
        """Start animation."""
        self._stop, self.is_running = False, True
        self.evt_active.emit(True)
        self.timer.singleShot(int(self.interval), self.advance)

    @contextmanager
    def pause(self):
        """Pause timer and then restart it."""
        if not self.is_running:
            raise ValueError("Cannot pause timer that is not running")
        self.timer.stop()
        yield
        self.timer.singleShot(int(self.interval), self.advance)

    def restart(self):
        """Restart animation."""
        self.finish()
        self.start()

    def finish(self):
        """Emit the finished event signal."""
        self._stop, self.is_running = True, False
        self.timer.stop()
        self.evt_active.emit(False)
        self.evt_finished.emit()

        if self._writer and self._writer_current_iter == 0:
            self._writer_finished = True
            self.evt_writer_finished.emit()

    def advance(self):
        """Advance the current frame in the animation.

        Takes dims scale into account and restricts the animation to the
        requested frame_range, if entered.
        """
        if self._stop:
            return

        self.current += self.step * self.dim_range[2]
        if self.current < self.min_point:
            self._writer_current_iter -= 1
            if self.loop_mode == LoopMode.BACK_AND_FORTH:  # 'loop_back_and_forth'
                self.step *= -1
                self.current = self.min_point + self.step * self.dim_range[2]
            elif self.loop_mode == LoopMode.LOOP:  # 'loop'
                self.current = self.max_point + self.current - self.min_point
            else:  # loop_mode == 'once'
                return self.finish()
        elif self.current >= self.max_point:
            self._writer_current_iter -= 1
            if self.loop_mode == LoopMode.BACK_AND_FORTH:  # 'loop_back_and_forth'
                self.step *= -1
                self.current = self.max_point + 2 * self.step * self.dim_range[2]
            elif self.loop_mode == LoopMode.LOOP:  # 'loop'
                self.current = self.min_point + self.current - self.max_point
            else:  # loop_mode == 'once'
                return self.finish()

        self.evt_next_requested.emit(self.current)
        # using a singleShot timer here instead of timer.start() because
        # it makes it easier to update the interval using signals/slots
        self.timer.singleShot(int(self.interval), self.advance)



"""Mixin classes for Qt widgets."""

import time
import typing as ty
from contextlib import contextmanager

import numpy as np
from koyo.timer import report_time
from loguru import logger
from qtpy.QtCore import QPoint, QRect, Qt, QTimer
from qtpy.QtGui import QCursor, QGuiApplication
from qtpy.QtWidgets import QApplication, QLayout, QWidget

import qtextra.helpers as hp


class ScreenManager:
    """Simple class that handles multi-screen logic."""

    def __init__(self):
        from qtpy.QtWidgets import QApplication

        self.screens = QApplication.screens()
        self.widths = [screen.geometry().width() for screen in self.screens]
        self.width = sum(self.widths)
        self.heights = [screen.geometry().height() for screen in self.screens]
        self.height = sum(self.heights)

    def get_minimum_size(self, width: int, height: int):
        """Get size that is suggested for current screen sizes."""
        self.widths.append(width)
        self.heights.append(height)
        return np.min(self.widths), np.min(self.heights)

    def verify_position(self, point: QPoint, width: int, height: int) -> QPoint:
        """Verify widget position is within the available geometry."""
        x_left, y_top = point.x(), point.y()
        # verify position horizontally
        if x_left < 0:
            x_left = 0
        x_right = x_left + width
        if x_right > self.width:
            x_right = self.width
            x_left = x_right - width
        # verify position vertically
        if y_top < 0:
            y_top = 0
        y_bottom = y_top - height
        if y_bottom > self.height:
            y_bottom = self.height
            y_top = y_bottom - height
        return QPoint(x_left, y_top)


class TimerMixin:
    """Timer mixin."""

    def _add_periodic_timer(self, interval: int, fcn, start: bool = True):
        """Create timer to execute some action."""
        timer = QTimer(self)
        timer.setInterval(interval)
        if fcn:
            timer.timeout.connect(fcn)

        if start:
            timer.start()
        logger.debug(f"Added periodic timer event that runs every {interval / 1000}s")
        return timer

    def _add_single_shot_timer(self, delay: int, fcn) -> QTimer:
        timer = QTimer(self)
        timer.singleShot(delay, fcn)
        return timer

    @contextmanager
    def measure_time(self, message: str = "Task took", func: ty.Callable = logger.trace):
        """Measure time."""
        t_start = time.time()
        yield
        func(f"{message} {report_time(t_start)}")


class MinimizeMixin:
    """Mixin class to enable hiding of popup."""

    def _make_hide_handle(self):
        hide_handle = hp.make_qta_btn(self, "minimise", tooltip="Click here to minimize the popup window")
        hide_handle.clicked.connect(self.on_hide)

        handle_layout = self._make_move_handle()
        handle_layout.addWidget(hide_handle)
        return hide_handle, handle_layout

    def on_hide(self):
        """Hide."""
        self.hide()
        self.clearFocus()

    def closeEvent(self, event):
        """Hide."""
        self.on_hide()
        event.ignore()


# noinspection PyUnresolvedReferences
class CloseMixin:
    """Mixin class to enable closing of popup."""

    HIDE_WHEN_CLOSE = False

    def _make_close_handle(self, title: str = ""):
        close_btn = hp.make_qta_btn(self, "cross", tooltip="Click here to close the popup window")
        close_btn.clicked.connect(self.close)

        handle_layout = self._make_move_handle()
        handle_layout.addWidget(close_btn)
        self._title_label.setText(title)
        return close_btn, handle_layout

    def _make_hide_handle(self, title: str = ""):
        self.HIDE_WHEN_CLOSE = True
        return self._make_close_handle(title)


class DialogMixin:
    """Mixin class for dialogs."""

    def show_above_widget(self, widget: QWidget, show: bool = True, y_offset: int = 14, x_offset: int = 0):
        """Show popup dialog above the widget."""
        rect = widget.rect()
        pos = widget.mapToGlobal(QPoint(rect.left() + rect.width() / 2, rect.top()))
        if show:
            self.show()
        sz_hint = self.size()
        pos -= QPoint((sz_hint.width() / 2) + x_offset, sz_hint.height() + y_offset)
        self.move(pos)

    def show_above_mouse(self, show: bool = True):
        """Show popup dialog above the mouse cursor position."""
        pos = QCursor().pos()  # mouse position
        sz_hint = self.sizeHint()
        pos -= QPoint(sz_hint.width() / 2, sz_hint.height() + 14)
        self.move(pos)
        if show:
            self.show()

    def show_below_widget(self, widget: QWidget, show: bool = True, y_offset: int = 14):
        """Show popup dialog above the widget."""
        rect = widget.rect()
        pos = widget.mapToGlobal(QPoint(rect.left() + rect.width() / 2, rect.top()))
        sz_hint = self.size()
        pos -= QPoint(sz_hint.width() / 2, -y_offset)
        self.move(pos)
        if show:
            self.show()

    def show_below_mouse(self, show: bool = True, x_offset: int = 0, y_offset: int = -14):
        """Show popup dialog above the mouse cursor position."""
        pos = QCursor().pos()  # mouse position
        sz_hint = self.sizeHint()
        pos -= QPoint(sz_hint.width() - x_offset / 2, y_offset)
        self.move(pos)
        if show:
            self.show()

    def show_right_of_widget(self, widget: QWidget, show: bool = True, x_offset: int = 14):
        """Show popup dialog above the widget."""
        rect = widget.rect()
        pos = widget.mapToGlobal(QPoint(rect.left() + rect.width() / 2, rect.top()))
        sz_hint = self.size()
        pos -= QPoint(-x_offset, sz_hint.height() / 4)
        self.move(pos)
        if show:
            self.show()

    def show_right_of_mouse(self, show: bool = True):
        """Show popup dialog on the right hand side of the mouse cursor position."""
        pos = QCursor().pos()  # mouse position
        sz_hint = self.sizeHint()
        pos -= QPoint(-14, sz_hint.height() / 4)
        self.move(pos)
        if show:
            self.show()

    def show_left_of_widget(self, widget: QWidget, show: bool = True, x_offset: int = 14):
        """Show popup dialog above the widget."""
        rect = widget.rect()
        pos = widget.mapToGlobal(QPoint(rect.left(), rect.top()))
        sz_hint = self.size()
        pos -= QPoint(sz_hint.width() + x_offset, sz_hint.height() / 4)
        self.move(pos)
        if show:
            self.show()

    def show_left_of_mouse(self, show: bool = True):
        """Show popup dialog on the left hand side of the mouse cursor position."""
        pos = QCursor().pos()  # mouse position
        sz_hint = self.sizeHint()
        pos -= QPoint(sz_hint.width() + 14, sz_hint.height() / 4)
        self.move(pos)
        if show:
            self.show()

    def set_on_widget(self, widget: QWidget, x_mult: float = 2.5, y_mult: float = 0.0):
        """Set position of the popup above the widget."""
        # calculate position information about the widget
        widget_rect = widget.rect()
        widget_pos = widget.mapToGlobal(QPoint(widget_rect.left(), widget_rect.top()))
        widget_width = widget.width()

        # calculate sizing of the window
        rect = self.rect()
        x_pos = widget_pos.x() + widget_width + rect.width() * x_mult
        y_pos = widget_pos.y() + rect.height() * y_mult
        pos = QPoint(x_pos, y_pos)

        m = ScreenManager()
        pos = m.verify_position(pos, rect.width(), rect.height())
        self.move(pos)

    def set_on_mouse(self, x_mult: float = 2.5, y_mult: float = 0.0):
        """Set on mouse position."""
        pos = QCursor.pos()
        rect = self.rect()
        pos = QPoint(pos.x() - rect.width() * x_mult, pos.y() - rect.height() * y_mult)

        m = ScreenManager()
        pos = m.verify_position(pos, rect.width(), rect.height())
        self.move(pos)

    def center_on_screen(self, show: bool = False):
        """Center dialog on screen."""
        screen = QApplication.desktop().screenGeometry()
        x = (screen.width() - self.width()) / 2
        y = (screen.height() - self.height()) / 2
        self.move(x, y)
        if show:
            self.show()

    def center_on_parent(self, show: bool = False):
        """Center on parent."""
        parent = self.parent()
        if not parent:
            self.center_on_screen()
        else:
            screen = parent.geometry()
            x = (screen.width() - self.width()) / 2
            y = (screen.height() - self.height()) / 2
            self.move(x, y)
        if show:
            self.show()

    def move_to(self, position="top", *, win_ratio=0.9, min_length=0):
        """Move popup to a position relative to the QMainWindow.

        Parameters
        ----------
        position : {str, tuple}, optional
            position in the QMainWindow to show the pop, by default 'top'
            if str: must be one of {'top', 'bottom', 'left', 'right' }
            if tuple: must be length 4 with (left, top, width, height)
        win_ratio : float, optional
            Fraction of the width (for position = top/bottom) or height (for
            position = left/right) of the QMainWindow that the popup will
            occupy.  Only valid when isinstance(position, str).
            by default 0.9
        min_length : int, optional
            Minimum size of the long dimension (width for top/bottom or
            height fort left/right).

        Raises
        ------
        ValueError
            if position is a string and not one of
            {'top', 'bottom', 'left', 'right' }
        """
        if isinstance(position, str):
            window = self.parent().window() if self.parent() else None
            if not window:
                raise ValueError(
                    "Specifying position as a string is only possible if the popup has a parent",
                )
            left = window.pos().x()
            top = window.pos().y()
            if position in ("top", "bottom"):
                width = window.width() * win_ratio
                width = max(width, min_length)
                left += (window.width() - width) / 2
                height = self.sizeHint().height()
                top += 24 if position == "top" else (window.height() - height - 12)
            elif position in ("left", "right"):
                height = window.height() * win_ratio
                height = max(height, min_length)
                # 22 is for the title bar
                top += 22 + (window.height() - height) / 2
                width = self.sizeHint().width()
                left += 12 if position == "left" else (window.width() - width - 12)
            else:
                raise ValueError(
                    'position must be one of ["top", "left", "bottom", "right"]',
                )
        elif isinstance(position, (tuple, list)):
            assert len(position) == 4, "`position` argument must have length 4"
            left, top, width, height = position
        else:
            raise ValueError(
                f"Wrong type of position {position}",
            )

        # necessary for transparent round corners
        self.resize(self.sizeHint())
        # make sure the popup is completely on the screen
        # In Qt ≥5.10 we can use screenAt to know which monitor the mouse is on

        if hasattr(QGuiApplication, "screenAt"):
            screen_geometry: QRect = QGuiApplication.screenAt(QCursor.pos()).geometry()
        else:
            # This widget is deprecated since Qt 5.11
            from qtpy.QtWidgets import QDesktopWidget

            screen_num = QDesktopWidget().screenNumber(QCursor.pos())
            screen_geometry = QGuiApplication.screens()[screen_num].geometry()

        left = max(min(screen_geometry.right() - width, left), screen_geometry.left())
        top = max(min(screen_geometry.bottom() - height, top), screen_geometry.top())
        self.setGeometry(left, top, width, height)

    def move_to_widget(self, widget: QWidget, position: str = "right"):
        """Move tutorial to specified widget."""
        x_pad, y_pad = 5, 5
        size = self.size()
        rect = widget.rect()
        if position == "left":
            x = rect.left() - size.width() - x_pad
            y = rect.center().y() - (size.height() * 0.5)
        elif position == "right":
            x = rect.right() + x_pad
            y = rect.center().y() - (size.height() * 0.5)
        elif position == "top":
            x = rect.center().x() - (size.width() * 0.5)
            y = rect.top() - size.height() - y_pad
        elif position == "bottom":
            x = rect.center().x() - (size.width() * 0.5)
            y = rect.bottom() + y_pad
        pos = widget.mapToGlobal(QPoint(x, y))
        self.move(pos)


class ScreenshotMixin:
    """Mixin class for taking screenshots."""

    @contextmanager
    def run_with_screenshot(self):
        """Must implement."""
        yield

    def _screenshot(self):
        return self.grab().toImage()

    def to_screenshot(self):
        """Get screenshot."""
        from napari._qt.dialogs.screenshot_dialog import ScreenshotDialog

        dialog = ScreenshotDialog(self.screenshot, self, history=[])
        if dialog.exec_():
            pass

    def screenshot(self, path: ty.Optional[str] = None):
        """Take screenshot of the viewer."""
        from napari._qt.utils import QImg2array

        with self.run_with_screenshot():
            img = self._screenshot()
        if path is not None:
            from skimage.io import imsave

            imsave(path, QImg2array(img))
        return QImg2array(img)

    def clipboard(self):
        """Take screenshot af the viewer and put it in the clipboard."""
        from qtextra.widgets.qt_button_clipboard import copy_image_to_clipboard

        with self.run_with_screenshot():
            img = self._screenshot()
        copy_image_to_clipboard(img)
        hp.add_flash_animation(self)

    def _get_save_screenshot_menu(self):
        """Get normalization menu."""
        menu = hp.make_menu(self)
        menu_save = hp.make_menu_item(self, "Save screenshot to file...", menu=menu)
        menu_save.triggered.connect(self.to_screenshot)
        menu_clip = hp.make_menu_item(self, "Copy screenshot to clipboard", menu=menu)
        menu_clip.triggered.connect(self.clipboard)
        return menu


class QtBase(TimerMixin, ScreenshotMixin):
    """Mixin class with common functionality for Dialogs and Tabs."""

    _main_layout = None

    DELAY_CONNECTION = False

    def __init__(self, parent=None, title: str = ""):
        self.logger = logger.bind(src=self.__class__.__name__)
        # Qt stuff
        if hasattr(self, "setWindowTitle"):
            self.setWindowTitle(str(title))
        if hasattr(self, "setAttribute"):
            self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        # Own attributes
        self._parent = parent
        # Make interface
        self.make_gui()
        # Connect signals
        if not self.DELAY_CONNECTION:
            self.connect_events()

    def make_panel(self) -> QLayout:
        """Make panel."""

    def make_gui(self):
        """Make and arrange main panel."""
        layout = self.make_panel()
        if layout is None:
            raise ValueError("Expected layout")
        if not layout.parent() and not self.layout():
            self.setLayout(layout)
        self._main_layout = layout

    def on_apply(self, *args):
        """Update config."""

    def on_teardown(self) -> None:
        """Teardown."""

    def connect_events(self, state: bool = True) -> None:
        """Connect events."""

    def closeEvent(self, event):
        """Hide rather than close."""
        self._on_teardown()
        self.connect_events(False)
        if hasattr(self, "evt_close"):
            self.evt_close.emit()
        self.close()
